
# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


import unittest
from .. import Samplers
from .. import exceptions
from .. import verbose

verbose.set_verbose_low()

class TestSequenceMetropolisHastings(unittest.TestCase):
    """Sequence of tests for the Metropolis Hastings sampler"""
    def setUp(self):
        pass
    
    def test_sample(self):
        """.."""
        pass
    

class TestSequenceGradientDescent(unittest.TestCase): 
    """Sequence of tests for the Gradient Descent sampler"""
    def setUp(self):
        pass
    
    def test_sample(self):
        """.."""
        pass



if __name__ == '__main__':
    unittest.main()







# ilang - Inference Language  
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# 20 Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging
# Dec 2013, Boston, MA, USA 


__all__ = ['Model','MultivariateGaussian','Poisson','Smoothness','Library']

from ilang.Graphs import Dependence 
from ilang.exceptions import *
from ilang.verbose import *  

from DisplayNode import DisplayNode
import numpy



class Model(Dependence,object): 
    """A Model is a Dependence with name. """ 
    def __init__(self, name):
        Dependence.__init__(self, name)





class MultivariateGaussian(Model): 
    variables          = {'x':'continuous','mu':'continuous','cov':'continuous'} 
    dependencies       = [['mu','x','directed'],['cov','x','directed']] 
    preferred_samplers = {'x':['HamiltonianMCMC']}

    # init
    def init(self): 
        pass 
         
    # expose to sampler: 
    def log_conditional_probability_x(self,x): 
        hessian = self._compute_hessian() 
        mu      = self.get_value('mu')
        return -.5*numpy.dot(numpy.dot((x-mu),hessian),(x-mu).T)

    def log_conditional_probability_gradient_x(self,x): 
        hessian = self._compute_hessian() 
        mu      = self.get_value('mu')        
        return -.5*numpy.dot((x-mu),hessian+hessian.T) 

    def log_conditional_probability_hessian_x(self,x): 
        hessian = self._compute_hessian()
        return hessian 

    def log_conditional_probability_mu(self,mu): 
        #FIXME: implement
        return 0 

    def log_conditional_probability_gradient_mu(self,mu): 
        #FIXME: implement
        return 0
        
    def sample_conditional_probability_x(self): 
        mu      = self.get_value('mu')
        cov     = self.get_value('cov')
        return numpy.random.multivariate_normal(mu.reshape(mu.size),cov) 

    # miscellaneous:  
    def _compute_hessian(self):
        cov     = self.get_value('cov')
        self._hessian = numpy.linalg.inv(cov) 
        return self._hessian 


class Poisson(Model): 
    variables = {'lambda':'continuous','alpha':'continuous','z':'discrete'} 
    dependencies = [['lambda','z','directed'],['alpha','z','directed']]
    preferred_samplers = {'lambda':['ExpectationMaximization'],'alpha':['ExpectationMaximization']}

    def init(self): 
        pass 

    def log_conditional_probability_lambda(self,Lambda): 
        return 0 

    def log_conditional_probability_gradient_lambda(self,Lambda): 
        return 0

    def log_conditional_probability_alpha(self,alpha): 
        return 0 

    def log_conditional_probability_gradient_alpha(self,alpha): 
        return 0
        
    def sample_conditional_probability_counts(self,counts): 
        return 0



class Smoothness(Model): 
    variables = {'x':'continuous','beta':'continuous'}
    dependencies = [['beta','x','directed']] 

    def init(self): 
        pass
        
    def log_conditional_probability_x(self,x): 
        return 0

    def log_conditional_probability_gradient_x(self,x): 
        return 0

    def log_conditional_probability_beta(self,beta): 
        return 0

    def log_conditional_probability_gradient_beta(self,beta): 
        return 0



class SSD_Transformation( Model ):
    variables = {'source':'continuous','target':'continuous','transformation':'continuous','sigma':'continuous'} 
    dependencies = [['source','target','directed'],['transformation','target','directed'],['sigma','target','directed']]

    def __init__( self, name=None ):
        if name  is None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
    
    def log_conditional_probability_transformation(self,transformation): 
        source = self.get_value('source') 
        target = self.get_value('target') 
        sigma  = self.get_value('sigma')
        log_p = 0.0
        return log_p

    def log_conditional_probability_gradient_transformation(self,transformation): 
        source = self.get_value('source') 
        target = self.get_value('target') 
        sigma  = self.get_value('sigma')
        gradient = numpy.asarray([0.0,0.0,0.0,0.0,0.0,0.0]) 
        return gradient 
        
    def sample_conditional_probability_target(self): 
        return 0    

    def init(self): 
        pass 






class Library(): 
    def __init__(self): 
        self.models = []
        self._make_inventory() 

    def list_models(self):  
        return self.models 
        
    def _make_inventory(self): 
        models = []   
        import inspect, sys
        for name,obj in inspect.getmembers(sys.modules[__name__], inspect.isclass):
            if issubclass(obj,Model): 
                if not obj == Model: 
                    models.append(obj) 
        self.models = models 
        return models 

    def export_dictionary(self):
        nodes = [] 
        links = [] 
        for model in self.list_models(): 
            d = model.export_dictionary() 
            model_nodes = d['nodes']
            model_links = d['links'] 
            for node in model_nodes: 
                node['name'] = model.get_name()+': '+node['name'] 
            for link in model_links: 
                link['source'] = model.get_name()+': '+link['source'] 
                link['target'] = model.get_name()+': '+link['target'] 
            nodes += model_nodes 
            links += model_links 
        dictionary = {'nodes':nodes,'links':links}
        return dictionary 

    def export_json(self):
        from json import dumps
        return dumps(self.export_dictionary()) 

    def display_in_browser(self):
        """Displays the graph in a web browser. """
        display_node = DisplayNode()
        display_node.display_in_browser('graph',self.export_dictionary()) 
     
    def _repr_html_(self): 
        display_node = DisplayNode()
        return display_node.display('graph',self.export_dictionary())._repr_html_()





# ilang - Inference Language  
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# 20 Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging
# Dec 2013, Boston, MA, USA 


__all__ = ['Node','Dependence','ProbabilisticGraphicalModel']

from ilang.exceptions import *
from ilang.verbose import * 
import numpy
from DisplayNode import DisplayNode 
        
data_types = ['continuous','discrete']

def set_data_type(self,node,data_type): 
    if data_type in data_types: 
        node.data_type = data_type
    else: 
        raise ParameterError("Unknown data_type '%s'. Known data types are %s."%(str(data_type),data_types)) 




def name(obj): 
    if hasattr(obj,'get_name'): 
        return obj.get_name()
    else: 
        return str(obj)
 
        
        
class NodeContainer(): 
    """Represents the node of a Dependence. It contains the information relative to a node, 
    necessary in order to define the Dependence. The NodeContainer is then attached to a 
    Node when the Dependence is associated to a ProbabilisticGraphicalModel. """
    
    def __init__(self, name): 
        self.name = name 
        self.parents = [] 
        self.children = [] 
        self.neighbours = [] 
        self.attached_node = None
        self.data_type = None

    def get_attached_node(self): 
        """Returns the name of the Node attached to the node container 
        (this can be None or a Node instance). """
        return self.attached_node 
    
    def get_name(self): 
        """Returns the name of the node container. This is defined in the Dependence object 
        of which the node container is part. (see Dependence - expecially the method 
        get_variables() ). """
        return self.name
    
    def has_attached_node(self): 
        """Returns True if there is a node in the node container. This can be True only if 
        the Dependence object of which the node container is part is attached to a 
        ProbabilisticGraphicalModel. """
        return isinstance(self.attached_node, Node)

    def set_data_type(self,data_type): 
        """Sets the data type associated to the node container. """
        set_data_type(self,data_type) 
        return True

    def attach_to_node(self,node): 
        """Attaches the node container to a node. This method is called by the Dependence object. 
        If the data types associated to the node and to the node container differ then if the data 
        type of one of the two is not defined, the data type is passed along. If the data types are 
        defined and discordant, the attachment fails. """
        # verify if the node and the container represent the same data type
        # if the data type of one of the two is not defined, share. 
        if not isinstance(node, Node): 
            raise UnexpectedParameterType("Expected an instance of Node") 
        if node.data_type == self.data_type: 
            self.__attach_to_node(node) 
        elif node.data_type == None: 
            node.set_data_type(self.data_type)
            self.__attach_to_node(node) 
        elif self.data_type == None: 
            self.set_data_type(node.data_type)
            self.__attach_to_node(node) 
        else:
            raise InconsistentGraph("Data type mismatch: node '%s' and node container '%s' have different data types ('%s' and '%s'). "%(repr(node),repr(self),node.data_type,self.data_type) )
        return True

    def __attach_to_node(self,node): 
        if self.has_attached_node(): 
            print_important("Detaching node container from node '%s' and attaching to node '%s'. \n"%(repr(self.attached_node),repr(node)) )
        self.attached_node = node

    def detach_from_node(self): 
        """Detaches the node container from the attached node. This method is called by the Dependence object. """
        self.attached_node = None




class Dependence(object): 
    def __init__(self,name): 
        self._define_instance_methods()     
        self.nodes_containers = []  
        self._make_nodes_containers() 
        self.init() 
        self.name = name
        
    def _make_nodes_containers(self): 
        """Instantiate a node container for each variable. """
        for variable_name in self.get_variables_names(): 
            nodeph = NodeContainer(variable_name) 
            self.nodes_containers.append(nodeph)

    def init(self): 
        """Subclass this method to define a new dependence. This method is called right 
        after the instantiation of the Dependence object, enabling the insertion of 
        initialization code. """
        return True 
      
    @classmethod
    def get_dependencies(cls): 
        """Returns a dictionary describing the dependencies between variables. """
        return cls.dependencies

    @classmethod
    def get_variables(cls): 
        """Returns a dictionary of the variables involved in the dependence object. """
        return cls.variables

    @classmethod
    def get_name(cls): 
        """Returns the name of the dependence. """
        if hasattr(cls,"name"): 
            return cls.name    
        elif hasattr(cls,"__name__"): 
            return cls.__name__
        elif hasattr(cls,"__class__"): 
            return cls.__class__.__name__ 
        else: 
            raise Exception("Something went wrong. This exception should never occur. ") 
        
    def attach_to_nodes(self,links): 
        """Attach the dependence object to the nodes of a ProbabilisticGraphicalModel. """
        for variable_name in links.keys(): 
            if not self.has_variable(variable_name): 
                raise ParameterError("Dependence does not have a variable with name '%s'. (The variables are %s). "%(variable_name,self.get_variables_names()))
            if not isinstance(links[variable_name], Node): 
                raise UnexpectedParameterType("Expected an instance of Node. ")
        for variable_name in links.keys(): 
            container = self.get_node_container(variable_name)
            container.attach_to_node(links[variable_name]) 

    def is_fully_attached(self): 
        """Returns True if all the variables are attached to the nodes of a graph. """
        pass 
        
    def get_nodes_containers(self): 
        """Returns a list of all the nodes containers. """
        return self.nodes_containers
      
    def get_node_container(self,variable_name): 
        """Returns the node container corresponding to the variable with the given name. """
        if not self.has_variable(variable_name): 
            raise InconsistentGraph("The requested variable (%s) does not exist. The variables are %s."%(variable_name,self.get_variables_names())) 
        for nodeph in self.get_nodes_containers(): 
            if nodeph.name == variable_name: 
                return nodeph  
         
    def get_node_from_variable_name(self,variable_name): 
        """Returns the node attached to the given variable. """
        if not self.has_variable(variable_name): 
            raise ParameterError("The requested variable (%s) does not exist. The variables are %s."%(variable_name,self.get_variables_names()))
        node_container = self.get_node_container(variable_name)
        if not node_container.has_attached_node(): 
            raise ParameterError("Variable '%s' is not attached to a node. "%variable_name)
        return node_container.get_attached_node() 
        
    def get_variable_name_from_node(self,node): 
        """Returns the name of the node container attached to the given node. """
        for container in self.get_nodes_containers(): 
            if isinstance(node,Node): 
                if container.get_attached_node() == node: 
                    return container.get_name() 
            elif isinstance(node,str): 
                if container.get_attached_node().name == node: 
                    return container.get_name() 
        raise ParameterError("node '%s' is not attached to any of the node containers. "%name(node)) 
        
    def get_variables_names(self): 
        """Returns a list of the names of all the variables. """
        return self.get_variables().keys()

    def get_attached_nodes(self): 
        """Returns a list of the nodes attached to the node containers. """
        # scan all node containers
        nodes = []
        for node_container in self.get_nodes_containers():
            node = node_container.get_attached_node()
            if isinstance(node, Node): 
                nodes.append(node)
        return nodes 
        
    def get_attached_nodes_names(self): 
        """Returns a list of the names of the nodes attached to the node containers. """
        names = []
        for node in self.get_attached_nodes(): 
            names.append(node.name)
        return names
        
    def get_attached_node(self,name): 
        """Returns the attached node with the given name. """
        for node in self.get_attached_nodes(): 
            if node.name == name: 
                return node
        return None
        
    def has_attached_node(self,node): 
        """Returns True if the given node (given by name or instance) is attached to one of the node containers. """
        for container in self.get_nodes_containers(): 
            if isinstance(node,Node): 
                if container.get_attached_node() == node: 
                    return True
            elif isinstance(node,str): 
                if container.get_attached_node().name == node: 
                    return True 
        return False 

    def set_name(self,name): 
        """Sets the name of the dependence object. """
        self.name = name

        
    # private
    def has_log_conditional_probability_gradient_variable(self,variable): 
        """Returns True if a method to compute the gradient of the log of the conditional 
        probability distribution associated to the given variable is defined. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        return hasattr(self,"log_conditional_probability_gradient_"+variable)
       
    def has_log_conditional_probability_hessian_variable(self,variable): 
        """Returns True if a method to compute the Hessian of the log of the conditional 
        probability distribution associated to the given variable is defined. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        return hasattr(self,"log_conditional_probability_hessian_"+variable)

    def has_log_conditional_probability_diagonal_hessian_variable(self,variable): 
        """Returns True if a method to compute the diagonal of the Hessian of the log of the conditional 
        probability distribution associated to the given variable is defined. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        return hasattr(self,"log_conditional_probability_diagonal_hessian_"+variable)
        
    def has_log_conditional_probability_variable(self,variable): 
        """Returns True if a method to compute the log of the conditional probability distribution 
        associated to the given variable is defined. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        return hasattr(self,"log_conditional_probability_"+variable) 

    def can_sample_conditional_probability_variable(self,variable): 
        """Returns True if a method to sample from the conditional probability of the given variable is defined. """ 
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        return hasattr(self,"sample_conditional_probability_"+variable)

    def get_log_conditional_probability_variable(self,variable,value=None): 
        """Returns the log of the conditional probability distribution 
        associated to the given variable. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        if not hasattr(self,"log_conditional_probability_"+variable): 
            raise ModelUndefined("The method to compute the log conditional probability of '%s' is not defined. "%variable) 
        if value is None: 
            value = self.get_value(variable)
        return eval("self.log_conditional_probability_"+variable+"(value)") #FIXME: eventually pass parameters

    def get_log_conditional_probability_gradient_variable(self,variable,value=None): 
        """Returns the gradient of the log of the conditional probability distribution 
        associated to the given variable. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        if not hasattr(self,"log_conditional_probability_gradient_"+variable): 
            raise ModelUndefined("The method to compute the gradient of the log conditional probability of '%s' is not defined. "%variable) 
        if value is None: 
            value = self.get_value(variable)
        return eval("self.log_conditional_probability_gradient_"+variable+"(value)") #FIXME: eventually pass parameters

    def get_log_conditional_probability_hessian_variable(self,variable,value=None): 
        """Returns the Hessian of the log of the conditional probability distribution 
        associated to the given variable. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        if not hasattr(self,"log_conditional_probability_hessian_"+variable): 
            raise ModelUndefined("The method to compute the Hessian of the log conditional probability of '%s' is not defined. "%variable) 
        if value is None: 
            value = self.get_value(variable)
        return eval("self.log_conditional_probability_hessian_"+variable+"(value)") #FIXME: eventually pass parameters
       
    def get_log_conditional_probability_diagonal_hessian_variable(self,variable,value=None): 
        """Returns the diagonal of the Hessian of the log of the conditional probability distribution 
        associated to the given variable. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        if not hasattr(self,"log_conditional_probability_hessian_"+variable): 
            raise ModelUndefined("The method to compute the diagonal Hessian of the log conditional probability of '%s' is not defined. "%variable) 
        if value is None: 
            value = self.get_value(variable)
        return eval("self.log_conditional_probability_diagonal_hessian_"+variable+"(value)") #FIXME: eventually pass parameters
        
    def sample_conditional_probability_variable(self,variable): 
        """Samples from the conditional probability distribution of the given variable. """
        if not self.has_variable(variable): 
            raise ParameterError("Variable '%s' does not exist. The variables are: %s. "%(variable,self.get_variables_names()))
        if not hasattr(self,"sample_conditional_probability_"+variable): 
            raise ModelUndefined("The method to sample from the log conditional probability of '%s' is not defined. "%variable) 
        return eval("self.sample_conditional_probability_"+variable+"()") #FIXME: eventually pass parameters
        
    def has_variable(self,variable): 
        return variable in self.variables.keys() 
        
    def get_value(self,variable):
        node = self.get_node_from_variable_name(variable)
        return node.get_value()
        
    # Inference interface 
    def has_log_conditional_probability_gradient_node(self,node): 
        """Returns True if a method to compute the gradient of the log of the conditional 
        probability distribution associated to the given node is defined. """
        if not self.has_node(node): 
            raise ParameterError("node '%s' does not exist. The nodes are: %s. "%(name(node),self.get_nodes_names()))        
        return self.has_log_conditional_probability_gradient_variable(self.get_variable_name_from_node(node)) 
       
    def has_log_conditional_probability_hessian_node(self,node): 
        """Returns True if a method to compute the Hessian of the log of the conditional 
        probability distribution associated to the given node is defined. """
        if not self.has_node(node): 
            raise ParameterError("node '%s' does not exist. The nodes are: %s. "%(name(node),self.get_nodes_names()))
        return self.has_log_conditional_probability_hessian_variable(self.get_variable_name_from_node(node)) 

    def has_log_conditional_probability_diagonal_hessian_node(self,node): 
        """Returns True if a method to compute the diagonal of the Hessian of the log of the conditional 
        probability distribution associated to the given node is defined. """
        if not self.has_node(node): 
            raise ParameterError("node '%s' does not exist. The nodes are: %s. "%(name(node),self.get_nodes_names()))
        return self.has_log_conditional_probability_diagonal_hessian_variable(self.get_variable_name_from_node(node)) 
        
    def has_log_conditional_probability_node(self,node): 
        """Returns True if a method to compute the log of the conditional probability distribution 
        associated to the given node is defined. """
        if not self.has_node(node): 
            raise ParameterError("node '%s' does not exist. The nodes are: %s. "%(name(node),self.get_nodes_names()))
        return self.has_log_conditional_probability_variable(self.get_variable_name_from_node(node)) 

    def can_sample_conditional_probability_node(self,node): 
        """Returns True if a method to sample from the conditional probability of the given node is defined. """ 
        if not self.has_node(node): 
            raise ParameterError("Variable '%s' does not exist. The nodes are: %s. "%(name(node),self.get_nodes_names()))
        if self.can_sample_conditional_probability_variable(self.get_variable_name_from_node(node)):  
            return True
        return False 

    def get_log_conditional_probability_node(self,node,value=None): 
        """Returns the log of the conditional probability distribution 
        associated to the given node (given by name or instance) - if possible. . """
        # get the name of the variable (node container) attached to the given node (if any). 
        variable_name = self.get_variable_name_from_node(node) 
        return self.get_log_conditional_probability_variable(variable_name,value)

    def get_log_conditional_probability_gradient_node(self,node,value=None): 
        """Returns the gradient of the log of the conditional probability distribution 
        associated to the given node (given by name or instance) - if possible. """
        # get the name of the variable (node container) attached to the given node (if any). 
        variable_name = self.get_variable_name_from_node(node) 
        return self.get_log_conditional_probability_gradient_variable(variable_name,value)

    def get_log_conditional_probability_hessian_node(self,node,value=None): 
        """Returns the Hessian of the log of the conditional probability distribution 
        associated to the given node (given by name or instance) - if possible. """
        # get the name of the variable (node container) attached to the given node (if any). 
        variable_name = self.get_variable_name_from_node(node) 
        return self.get_log_conditional_probability_hessian_variable(variable_name,value)

    def get_log_conditional_probability_diagonal_hessian_node(self,node,value=None): 
        """Returns the diagonal of the Hessian of the log of the conditional probability distribution 
        associated to the given node (given by name or instance) - if possible. """
        # get the name of the variable (node container) attached to the given node (if any). 
        variable_name = self.get_variable_name_from_node(node) 
        return self.get_log_conditional_probability_diagonal_hessian_variable(variable_name,value)
        
    def sample_conditional_probability_node(self,node):
        """Sample from the given node - if possible). """
        # get the name of the variable (node container) attached to the given node (if any). 
        variable_name = self.get_variable_name_from_node(node) 
        return self.sample_conditional_probability_variable(variable_name)

    def has_node(self,node): 
        for container in self.get_nodes_containers(): 
            if isinstance(node,Node): 
                if container.get_attached_node() == node: 
                    return True
            elif isinstance(node,str): 
                if container.get_attached_node().name == node: 
                    return True
        return False 
                    
    @classmethod
    def export_pickle(cls): 
        """Export a pickled representation of the graph. """
        from pickle import dumps
        return dumps(cls.export_dictionary())
 
    @classmethod
    def export_json(cls): 
        """Export a json representation of the graph. """
        from json import dumps
        return dumps(cls.export_dictionary())


    # Export graph  
    @classmethod
    def export_dictionary(cls): 
        """Export a dictionary that describes the conditional independencies. """
        graph = {}
        vars = cls.get_variables()
        deps = cls.get_dependencies() 
        # list the nodes 
        nodes = []
        for var_name in vars.keys(): 
            if vars[var_name] == 'continuous': 
                node_type = 4
            else: 
                node_type = 5
            nodes.append({'name':var_name,'type':node_type})
        graph['nodes'] = nodes
        # list the dependencies 
        links = []
        for dep in deps: 
                link_source = dep[0]
                link_target = dep[1]
                if dep[2] == 'directed': 
                    link_type = 't1'
                else: 
                    link_type = 't2'
                links.append({'source':link_source,'target':link_target,'type':link_type}) 
        graph['links'] = links
        return graph

    # Visualisation 
    @classmethod
    def display_in_browser(cls): 
        """Displays the graph in a web browser. """
        display_node = DisplayNode()
        display_node.display_in_browser('graph',cls.export_dictionary()) 

    # Note: as of Dec. 2013, IPython does not find _repr_html_ if it is a class method, therefore use .display() to display 
    # the graph of a model derived from Dependence: e.g. (in Ipython): 
    # from Model import MultivariateGaussian
    # MultivariateGaussian.display() 
    @classmethod
    def _repr_html_(cls): 
        display_node = DisplayNode()
        return display_node.display('graph',cls.export_dictionary())._repr_html_()
        
    # Note: as of Dec. 2013, IPython does not find _repr_html_ if it is a class method, therefore use .display() to display 
    # the graph of a model derived from Dependence: e.g. (in Ipython): 
    # from Model import MultivariateGaussian
    # MultivariateGaussian.display() 
    @classmethod
    def display(cls): 
        display_node = DisplayNode()
        return display_node.display('graph',cls.export_dictionary())
                
    def _define_instance_methods(self): 
        def get_dependencies(): 
            """Returns a dictionary describing the dependencies between variables. """
            return self.dependencies
        setattr(self,'get_dependencies',get_dependencies)
        
        def get_variables(): 
            """Returns a dictionary of the variables involved in the dependence object. """
            return self.variables    
        setattr(self,'get_variables',get_variables)

        def get_name(): 
            if hasattr(self,"name"): 
                return self.name    
            elif hasattr(self,"__name__"): 
                return self.__name__
            elif hasattr(self,"__class__"): 
                return self.__class__.__name__ 
            else: 
                raise Exception("Something went wrong. This exception should never occur. ") 
        setattr(self,'get_name',get_name)
        
        def export_dictionary(): 
            """Export a dictionary that describes the conditional independencies. """
            graph = {}
            vars = self.get_variables()
            deps = self.get_dependencies() 
            # list the nodes 
            nodes = []
            for var_name in vars.keys(): 
                if vars[var_name] == 'continuous': 
                    node_type = 4
                else: 
                    node_type = 5
                nodes.append({'name':var_name,'type':node_type})
            graph['nodes'] = nodes
            # list the dependencies 
            links = []
            for dep in deps: 
                    link_source = dep[0]
                    link_target = dep[1]
                    if dep[2] == 'directed': 
                        link_type = 't1'
                    else: 
                        link_type = 't2'
                    links.append({'source':link_source,'target':link_target,'type':link_type}) 
            graph['links'] = links
            return graph
        setattr(self,'export_dictionary',export_dictionary)
        
        def export_pickle(): 
            """Export a pickled representation of the graph. """
            from pickle import dumps
            return dumps(self.export_dictionary())
        setattr(self,'export_pickle',export_pickle)
         
        def export_json(): 
            """Export a json representation of the graph. """
            from json import dumps
            return dumps(self.export_dictionary())
        setattr(self,'export_json',export_json)
        
        def display_in_browser(): 
            """Displays the graph in a web browser. """
            display_node = DisplayNode()
            display_node.display_in_browser('graph',self.export_dictionary()) 
        setattr(self,'display_in_browser',display_in_browser)
        
        def _repr_html_(cls): 
            display_node = DisplayNode()
            return display_node.display('graph',cls.export_dictionary())._repr_html_()
        setattr(self.__class__,'_repr_html_',_repr_html_)

        

        
            
class Node(): 
    """Node of a graphical model. A node represents a variable or set of variables. 
    The node is characterized by name, value, and by the given flag. The flag is True if 
    the variable is considered a given quantity, False if it considered unknown (uncertain). """
    def __init__(self,name,value=None,given=False): 
        self.name = name
        self.value = value
        self.given = bool(given)
        self.dependencies_attached = []
        self.data_type = None
        
    def set_given(self,is_given): 
        """Sets the state of the given flag. """
        self.given = bool(is_given)
        return True
        
    def is_given(self):
        """Returns the state of the given flag""" 
        return self.given
        
    def set_value(self,value):
        """Sets the value associated to the node. The value must be an instance of numpy.ndarray"""
        if not (isinstance(value,numpy.ndarray) or numpy.isscalar(value)):  
            try: 
                value = numpy.asarray(value)
            except: 
                raise UnexpectedParameterType("'value' is expected to be an instance of numpy.ndarray") 
        self.value = value
#        # set value, but maintain the object in the same memory location, so that all memory references remain valid 
#        print "@@ Setting value: ",self.name, value
#        if self.value is None: 
#            if numpy.isscalar(value):
#                self.value = numpy.asarray(value)
#            else: 
#                self.value = value
#        else: 
#            if numpy.isscalar(value): 
#                self.value.data = numpy.asarray(value).data
#            else: 
#                self.value[:] = value
        return True
        
    def set_name(self,name): 
        """Sets the name of the node. """
        self.name = str(name)
        return True

    def get_value(self):
        """Returns the value associated to the node. """
        return self.value
        
    def get_name(self):
        """Returns the name of the node. """
        return self.name
    
    def get_dependencies_attached(self): 
        """Returns a list of the Dependence objects associated to the node 
        (it's an empty list if the node is not inserted in a Probabilistic Graphical Model). """    
        return self.dependencies_attached

    def set_data_type(self,data_type): 
        """Sets the data type associated with the node. """
        set_data_type(self,data_type) 
        return True
    
    def zeros(self): 
        """Return an array of zeros of the same size and shape as the value of the node. """
        return numpy.zeros(self.value.shape)

    def zeros_hessian(self):
        """Return a square matrix of zeros, with number of rows equal to the length of the value of the node. """
        return numpy.zeros((self.value.size,self.value.size))

    def has_log_conditional_probability_gradient(self): 
        """Returns True if a method to compute the gradient of the log of the conditional 
        probability distribution is defined. """
        for dependence in self.get_dependencies_attached(): 
           if not dependence.self.has_log_conditional_probability_gradient_node(self): 
               return False
        return True 
       
    def has_log_conditional_probability_hessian(self): 
        """Returns True if a method to compute the Hessian of the log of the conditional 
        probability distribution is defined. """
        for dependence in self.get_dependencies_attached(): 
           if not dependence.self.has_log_conditional_probability_hessian_node(self): 
               return False
        return True 

    def has_log_conditional_probability_diagonal_hessian(self): 
        """Returns True if a method to compute the diagonal of the Hessian of the log of the conditional 
        probability distribution is defined. """
        for dependence in self.get_dependencies_attached(): 
           if not dependence.self.has_log_conditional_probability_diagonal_hessian_node(self): 
               return False
        return True 
        
    def has_log_conditional_probability(self): 
        """Returns True if a method to compute the log of the conditional probability distribution is defined. """
        for dependence in self.get_dependencies_attached(): 
           if not dependence.self.has_log_conditional_probability_node(self): 
               return False
        return True 

    def can_sample_conditional_probability(self): 
        """Returns True if a method to sample from the conditional probability is defined. """ 
        for dependence in self.get_dependencies_attached(): 
           if not dependence.can_sample_conditional_probability_node(): 
               return False
        return True 
        

        
class ProbabilisticGraphicalModel():
    """Probabilistic Graphical Model. This is the central object of ilang. It represents a probabilistic 
    model. A specific model is constructed by setting Nodes and Dependencies. """
    def __init__(self,node_names=[]):
        if not isinstance(node_names,type([])): 
            raise UnexpectedParameterType("Optional list_of_names must be a list. ")
        self.nodes = {}  
        self.dependencies = {} 
        self.add_nodes(node_names) 
        
    # Handling of nodes
    def add_node(self,node,value=None,given=None): 
        "Add single node to the Probabilistic Graphical Model. Optionally specify value and given flag" 
        # check parameter type:  
        if not (isinstance(node,Node) or isinstance(node,str)):
            raise UnexpectedParameterType("node '%s' is not an instance of Node or a string identifying the name of a new node"%name(node)) 
        # check if node is already in the dag
        if isinstance(node,str): 
            name = node
        else: 
            name = node.name
        if self.has_node(node) or self.has_node(name): 
            raise ParameterError("The graphical model already has the given node (node name: '%s') "%str(name))
        # modify the node value and flag if specified: 
        if isinstance(node,Node): 
            if value  is not None:
                node.set_value(value)
            if given  is not None: 
                node.set_given(given)
        else: 
            node = Node(name,value,given)
        self.nodes[name] = node
        return True
            
        
    def add_nodes(self,list_of_nodes):
        """Add multiple nodes to the Probabilistic Graphical Model. """
        # check for correct parameter type
        if not isinstance(list_of_nodes,type([])): 
            list_of_nodes = [list_of_nodes,]
        for node in list_of_nodes: 
            if not (isinstance(node,Node) or isinstance(node,str)): 
                raise UnexpectedParameterType("node % is not an instance of Node or a string identifying the name of a new node. "%name(node))            
        names = [] 
        # verify if the given names are unique
        for node in list_of_nodes: 
            if isinstance(node,str): 
                names.append(node)
            else: 
                names.append(node.name)
        if len(names)!=len(set(names)): 
            raise ParameterError("There are nodes with identical names in the list. Each node of a Probabilistic Graphical Model must have a unique name. ")
        # verify if the nodes are unique (this applies in case nodes are Node instances)
        if len(list_of_nodes)!=len(set(list_of_nodes)): 
            raise ParameterError("There are identical nodes in the list. Each node must appear only once in the graph. ")
        # check if any of the given nodes or node names already exist 
        for node in list_of_nodes: 
            if isinstance(node,str): 
                name = node
            else: 
                name = node.name
            # check if node with same name exists
            if self.has_node(node) or self.has_node(name): 
                raise ParameterError("The graphical model already has the given node (node name: '%s') "%str(name))
        # add the nodes to the graph
        for node in list_of_nodes: 
            self.add_node(node)
        return True
 
    def get_node(self,node): 
        """Returns the node with given name. """
        if not self.has_node(node): 
            raise ParameterError("Node with name '%s' does not exist. "%name(node))
        if isinstance(node,Node): 
            return node
        else: 
            return self.nodes[node]
        
    def get_nodes(self): 
        """Returns a list of all nodes. """
        return self.nodes.values() 
        
    def get_nodes_names(self): 
        """Returns a list of the names of all nodes. """
        return self.nodes.keys() 

    def is_node_given(self,node): 
        """Returns True is the node represents a given quantity. See the properties of Node. """
        if isinstance(node,Node): 
            name = node.name 
        elif isinstance(node,str): 
            name = node
        else: 
            raise UnexpectedParameterType("node must be an instance of Node or a valid node name. ")
        node = self.get_node(name) 
        if not isinstance(node,Node): 
            raise ParameterError ("Probabilistic Graphical Model does not have a node named "+str(name))
        return node.is_given() 
        
    def set_nodes_given(self,list_of_nodes,is_given): 
        """Sets the state of the 'given' flag of the nodes in the list. """
        if not isinstance(list_of_nodes,type([])): 
            list_of_nodes = [list_of_nodes,]
        for node in list_of_nodes: 
            if isinstance(node,Node): 
                name = node.name 
            elif isinstance(node,str): 
                name = node
            else: 
                raise UnexpectedParameterType("node must be an instance of Node or a valid node name. ")
            node = self.get_node(name) 
            if not isinstance(node,Node): 
                raise ParameterError ("Probabilistic Graphical Model does not have a node named "+str(name))
        for node in list_of_nodes: 
            node = self.get_node(node) 
            node.set_given(is_given) 
        return True 
        
    def get_node_value(self, node): 
        """Returns the value of the node. """
        if isinstance(node,Node): 
            name = node.name 
        elif isinstance(node,str): 
            name = node
        else: 
            raise UnexpectedParameterType("node must be an instance of Node or a valid node name. ")
        node = self.get_node(name) 
        if not isinstance(node,Node): 
            raise ParameterError ("Probabilistic Graphical Model does not have a node named "+str(name))
        return node.get_value() 
            
    def set_node_value(self, node, value): 
        """Sets the value of the node. """
        if isinstance(node,Node): 
            name = node.name 
        elif isinstance(node,str): 
            name = node
        else: 
            raise UnexpectedParameterType("node must be an instance of Node or a valid node name. ")
        node = self.get_node(name) 
        if not isinstance(node,Node): 
            raise ParameterError ("Probabilistic Graphical Model does not have a node named "+str(name))
        node.set_value(value) 
        return True 

    def has_node(self,node):
        """Returns True if the Probabilistic Graphical Model includes a given node (given by name or node instance). """
        if isinstance(node,Node): 
            name = node.name
        else: 
            name = node 
        return name in self.nodes.keys() 

    def remove_node(self,node): 
        """Removes from the Probabilistic Graphical Model the given node (given by name or node instance)"""
        if isinstance(node,Node): 
            name = node.name
        else: 
            name = node    
        if not self.has_node(node): 
            raise ParameterError ("Probabilistic Graphical Model does not have a node named "+str(name))      
        return self.nodes.pop(name)

    def list_nodes(self): 
        return self.nodes.keys() 

        
    # Handling of dependencies
    def add_dependence(self,dependence,links=None): 
        """Adds a probabilistic dependence between nodes (or amongst the variables of a single - milti-dimensional - node). 
        'dependence' must be an instance of Dependence. 'links' is optional and specifies how the variables of object 'dependence' 
        must be attached to the nodes. 'attach_dependence' can be used alternatively (at a later stage).  """ 
        # check for correct parameter type
        if not isinstance(dependence,Dependence): 
            raise UnexpectedParameterType("dependence must be an instance of Dependence")
        if links  is not None:
            if not isinstance(links, type({})):
                raise UnexpectedParameterType("links must be a dictionary") 
        # check if the dependence has been already added to the graph  
        if self.has_dependence(dependence): 
            raise ParameterError("The graphical model already has the dependence '%s'. "%str(dependence.get_name()))
        # add the dependence 
        self.dependencies[dependence.get_name()] = dependence 
        # optionally attach dependence
        if links  is not None: 
            self.attach_dependence_to_nodes(dependence,links)
        return True

    def remove_dependence(self,dependence): 
        """Removes the dependence (given by name or instance) from the Probabilistic Graphical. """
        if isinstance(dependence,Dependence): 
            name = dependence.get_name()
        else: 
            name = dependence  
        if not self.has_dependence(dependence): 
            raise ParameterError ("The dependence '%s' is not attached to the Probabilistic Graphical Model. "+name)      
        #FIXME: tell the nodes and tell the dependence (returned by the current function, though detached)  
        return self.dependencies.pop(name)
        
    def get_dependence(self,name): 
        """Returns the dependence with given name. """
        if not self.has_dependence(name): 
            raise ParameterError("Dependence with name '%s' does not exist. "%name)
        return self.dependencies[name]

    def get_dependencies(self): 
        """Returns a list of all the dependencies attached to the graph. """
        return self.dependencies.values() 

    def get_dependencies_names(self): 
        """Returns a list of the names of all the dependencies attached to the graph. """
        return self.dependencies.keys() 

    def has_dependence(self,dependence):
        """Returns True if the Probabilistic Graphical Model includes the dependence (given by name or instance). """
        if isinstance(dependence,Dependence): 
            name = dependence.get_name()
        else: 
            name = dependence
        return name in self.dependencies.keys() 

    def attach_dependence_to_nodes(self,dependence,links): 
        """Attach the variables of the given dependence to the nodes of the graph. 
        The links between variables and nodes must be specified in a dictionary: 
        {'variable_name1':'node_name1','variable_name2':'node_name2'} """
        # make sure that the dependence object is one of the dependencies of the graph
        if not self.has_dependence(dependence): 
            raise ParameterError("The specified dependence is not one of the dependencies of the graph. Add the dependence first: add_dependence()") 
        # verify that all the variables specified in the links dictionary exist
        for variable_name in links.keys(): 
            if not dependence.has_variable(variable_name): 
                raise ParameterError("One of the specified variables (%s) does not exist. The variables are %s. "%(variable_name,dependence.get_variables_names()))
        # verify that all the nodes specified in the links dictionary exist
        for node in links.values(): 
            if not self.has_node(node): 
                raise ParameterError("One of the specified nodes (%s) does not exist. The nodes are %s. "%(node,self.get_nodes_names()))
        # replace nodes names with nodes if nodes were specified by name 
        for variable_name in links.keys(): 
            node = links[variable_name]
            if isinstance(node,str): 
                node = self.get_node(node) 
                links[variable_name] = node
        # tell dependence to which nodes it is now attached (dependence will tell each of the nodes about it) 
        dependence.attach_to_nodes(links) 

    def is_dependence_fully_attached(self,dependence): 
        """Returns True if all the variables of the dependence object are attached to a node. """
        return dependence.is_fully_attached() 


    # Properties of the graph
    def d_separation(self, nodeA, nodeB, given_node): 
        """Checks if nodeA is independent from nodeB given node given_node. """
        pass
    
    def markov_blanket(self, node): 
        """Return the list of nodes that form the Markov Blanket of the given node. """
        pass 

    def get_parents(self, node): 
        """Returns the list of parents of the given node. """
        pass
            
    def get_children(self, node): 
        """Returns the list of children of the given node. """
        pass 
        
    def get_node_dependencies(self, node): 
        """Returns the list of dependencies attached to the given node. """
        # scan all the dependencies of the probabilistic graphical model 
        dependencies = []
        for d in self.get_dependencies():
            if d.has_attached_node(node): 
                dependencies.append(d)
        return dependencies 
                       
    def is_complete(self): 
        """Returns True if for every dependence attached to the graph all variables are attached to a node. 
        Therefore it returns False if there are variables of at least of Dependence object not attached to a node of the graph. """
        # check if all the attached dependencies are fully attached: 
        for dependence in self.get_dependencies(): 
            if not self.is_dependence_fully_attached(dependence): 
                return False
        # verify if there are unattached nodes: 
        for node in self.get_nodes(): 
            if self.get_node_dependencies(node) == []: 
                return False
        return True 



    # Inference
    def get_log_conditional_probability(self,node,value=None): 
        """Returns the log of the conditional probability of the given node. """
        # sum over the dependencies associated to the given node 
        log_p = 0 
        for dependence in self.get_node_dependencies(node): 
            if not dependence.has_log_conditional_probability_node(node): 
                raise UndefinedModelProperty("Dependence '%s' does not have a method to compute the log probability of the conditional probability of '%s' ( -> '%s'). "%(dependence.get_name(),name(node),dependence.get_variable_name_from_node(node)))
        for dependence in self.get_node_dependencies(node): 
            log_p += dependence.get_log_conditional_probability_node(node,value)
        return log_p

    def get_log_conditional_probability_gradient(self,node,value=None): 
        """Returns the gradient of the log of the conditional probability of the given node. """
        # sum over the dependencies associated to the given node 
        node = self.get_node(node)
        gradient = node.zeros() 
        print_debug("  summing over dependencies: ")
        for dependence in self.get_node_dependencies(node): 
            print_debug("  - dependence: %s"%name(dependence))
            if not dependence.has_log_conditional_probability_gradient_node(node): 
                raise UndefinedModelProperty("Dependence '%s' does not have a method to compute the gradient of the log probability of the conditional probability of '%s' ( -> '%s'). "%(dependence.get_name(),name(node),dependence.get_variable_name_from_node(node)))
        for dependence in self.get_node_dependencies(node): 
            gradient += (dependence.get_log_conditional_probability_gradient_node(node,value)).reshape(gradient.shape)
        return gradient 
        
    def get_log_conditional_probability_hessian(self,node,value=None): 
        """Returns the Hessian of the log of the conditional probability of the given node. """
        # sum over the dependencies associated to the given node 
        node = self.get_node(node)
        hessian = node.zeros_hessian() 
        for dependence in self.get_node_dependencies(node): 
            if not dependence.has_log_conditional_probability_hessian_node(node): 
                raise UndefinedModelProperty("Dependence '%s' does not have a method to compute the Hessian of the log probability of the conditional probability of '%s' ( -> '%s'). "%(dependence.get_name(),name(node),dependence.get_variable_name_from_node(node)))
        for dependence in self.get_node_dependencies(node): 
            hessian += (dependence.get_log_conditional_probability_hessian_node(node,value)).reshape(hessian.shape)
        return hessian 

    def get_log_conditional_probability_diagonal_hessian(self,node,value=None): 
        """Returns the Hessian of the log of the conditional probability of the given node. """
        # sum over the dependencies associated to the given node 
        node = self.get_node(node)
        diaghessian = node.zeros() 
        for dependence in self.get_node_dependencies(node): 
            if not dependence.has_log_conditional_probability_diagonal_hessian_node(node): 
                raise UndefinedModelProperty("Dependence '%s' does not have a method to compute the diagonal of the Hessian of the log probability of the conditional probability of '%s' ( -> '%s'). "%(dependence.get_name(),name(node),dependence.get_variable_name_from_node(node)))
        for dependence in self.get_node_dependencies(node): 
            diaghessian += (dependence.get_log_conditional_probability_diagonal_hessian_node(node,value)).reshape(diaghessian.shape)
        return diaghessian 

    def can_sample_conditional_probability_node(self,node): 
        if len(self.get_node_dependencies(node)): 
            dependence = self.get_node_dependencies(node)[0]
            if dependence.can_sample_conditional_probability_node(node): 
                return True 
        return False 

    def sample_conditional_probability_node(self,node): 
        #direct sampling - just calls the function at the node (only if the node has just one parent)
        #(self.can_sample_conditional_probability_node() returns True only if the node has only one parent) 
        if not self.can_sample_conditional_probability_node(node): 
            raise NoCompatibleSampler("Node '%s' cannot be sampled by direct sampling. "%(name(node)))
        dependence = self.get_node_dependencies(node)[0]  #the list has length 1 if self.can_sample_conditional_probability_node() returns True
        return dependence.sample_conditional_probability_node(node) 
                
    # Export graph    
    def export_dictionary(self): 
        """Export a dictionary that describes the graph. """
        graph = {}
        # list the nodes 
        nodes = []
        for node in self.get_nodes(): 
            if node.is_given(): 
                node_type = 1
            else: 
                node_type = 0
            nodes.append({'name':node.name,'type':node_type})
        graph['nodes'] = nodes
        # walk through the dependence objects and list all the (inner) dependencies 
        links = []
        for dependence in self.get_dependencies(): 
            dependencies = dependence.get_dependencies() 
            for link in dependencies: 
                if link[2] == 'directed': 
                    link_type = 't1'
                else: 
                    link_type = 't2'
                link_source = dependence.get_node_from_variable_name(link[0]).name
                link_target = dependence.get_node_from_variable_name(link[1]).name
                links.append({'source':link_source,'target':link_target,'type':link_type}) 
        graph['links'] = links
        return graph

    def export_pickle(self): 
        """Export a pickled representation of the graph. """
        from pickle import dumps
        return dumps(self.export_dictionary())
 
    def export_json(self): 
        """Export a json representation of the graph. """
        from json import dumps
        return dumps(self.export_dictionary())


    # Visualisation 
    def display_in_browser(self):
        """Displays the graph in a web browser. """
        display_node = DisplayNode()
        display_node.display_in_browser('graph',self.export_dictionary()) 
     
    def _repr_html_(self): 
        display_node = DisplayNode()
        return display_node.display('graph',self.export_dictionary())._repr_html_()

 
 



# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


__all__ = ['FileSource']

from ilang.Models import * 
from ilang.exceptions import *
from ilang.verbose import * 
import nibabel


class FileSource(object):
    def __init__(self,filename=None):
        self.nifti_image=None
        if filename != None: 
            self.load(filename)

    def load(self,file): 
        self.nifti_image = nibabel.load(file) 

    def get_data(self): 
        return self.nifti_image.get_data() 
 
    def get_affine(self):
        return self.nifti_image.get_affine()
        
    def get_info(self): 
        return self.nifti_image.get_info() 

    def save(self,file):
        return self.nifti_image.save(file)  
 
    def get_data_dtype(self):
        return self.nifti_image.get_data_type() 
         
    def get_filename(self):
        return self.nifti_image.get_filename() 
         
    def get_header(self):
        return self.nifti_image.get_header()  
        
    def get_qform(self):
        return self.nifti_image.get_qform()  
        
    def get_sform(self):
        return self.nifti_image.get_sform()  
        
    def get_shape(self):
        return self.nifti_image.get_shape()  

    def set_filename(self,filename):
        return self.nifti_image.set_filename(filename) 
        
    def set_qform(self,q):
        return self.nifti_image.set_qform(q)
        
    def set_sform(self,s):
        return self.nifti_image.set_sform(s)
        
    def is_instance_to_filename(self):
        return self.nifti_image.is_instance_to_filename()
        
    # display
    def show(self): 
        from math import floor 
        from numpy import zeros
        import Image
        from webgui.QuickDisplay import display_viewport3
        imx = self.get_data()[floor(self.get_shape()[0]/2),:,:]
        imy = self.get_data()[:,floor(self.get_shape()[1]/2),:]
        if len(self.get_shape())>=3: #FIXME: handle 4D, 5D, ..
            imz = self.get_data()[:,:,floor(self.get_shape()[2]/2)]
        else: 
            imz = zeros((128,128))
        imx = Image.fromarray(imx).rotate(90)
        imy = Image.fromarray(imy).rotate(90)
        imz = Image.fromarray(imz).rotate(90)
        return display_viewport3(imx,imy,imz,background=True,new_tab=True)
        
    def _repr_html_(self): 
        from math import floor 
        from numpy import zeros
        import Image
        from webgui.QuickDisplay import display_viewport3_ipython_notebook
        imx = self.get_data()[floor(self.get_shape()[0]/2),:,:]
        imy = self.get_data()[:,floor(self.get_shape()[1]/2),:]
        if len(self.get_shape())>=3: #FIXME: handle 4D, 5D, ..
            imz = self.get_data()[:,:,floor(self.get_shape()[2]/2)]
        else: 
            imz = zeros((128,128))
        imx = Image.fromarray(imx).rotate(90)
        imy = Image.fromarray(imy).rotate(90)
        imz = Image.fromarray(imz).rotate(90)
        return display_viewport3_ipython_notebook(imx,imy,imz)








__all__ = ['ProgressBar','LIGHT_BLUE','LIGHT_RED','BLUE','RED','GRAY','LIGHT_GRAY']

import uuid
from IPython.display import HTML, Javascript, display


LIGHT_BLUE   = 'rgb(200,228,246)'
BLUE         = 'rgb(47,128,246)'
LIGHT_RED    = 'rgb(246,228,200)'
RED          = 'rgb(246,128,47)'
LIGHT_GRAY   = 'rgb(246,246,246)'
GRAY         = 'rgb(200,200,200)'

    
class ProgressBar(): 
    def __init__(self, height='6px', width='100%%', background_color=LIGHT_BLUE, foreground_color=BLUE): 
        self.divid = str(uuid.uuid4())
        self.pb = HTML(
        """
        <div style="border: 1px solid white; width:%s; height:%s; background-color:%s">
            <div id="%s" style="background-color:%s; width:0%%; height:%s"> </div>
        </div> 
        """ % ( width, height, background_color, self.divid, foreground_color, height))
        self.visible = False
        self._previous = -1.0 

    def show(self):
        display(self.pb)
        self.visible = True 
        
    def set_percentage(self,percentage):
        if not self.visible: 
            self.show()
        if percentage < 1:
            percentage = 1
        if percentage > 100:
            percentage = 100 
        percentage = int(percentage)
        if percentage != self._previous:
            self._previous = percentage
            display(Javascript("$('div#%s').width('%i%%')" % (self.divid, percentage)))




# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Dec 2013, Boston 


__all__ = ['Sampler'] 

from ilang.Graphs import Node, ProbabilisticGraphicalModel, name
from ilang.Tracers import RamTracer
from ilang.exceptions import * 
from ilang.verbose import * 
from ilang.accessories import ProgressBar, LIGHT_RED, RED

import numpy
import inspect, sys
import ipy_table 
try:
    import scipy
except: 
    print "Please install Scipy"
    has_scipy = False
else:
    has_scipy = True
    from scipy.optimize import fmin_l_bfgs_b


MAX_RANDOM_NODES = 10000




# Sampling strategies 
class SamplingStrategy(object): 
    def __init__(self,graph): 
        self.graph = graph 

    def next_node(self): 
        pass 

        
class RandomNodesStrategy(SamplingStrategy): 
    def __init__(self, *args, **kwds):
        super(RandomNodesStrategy, self).__init__(*args, **kwds)       

    def next_node(self): 
        nodes = self.graph.get_nodes()
        node = None
        for i in range(MAX_RANDOM_NODES): 
            index = numpy.random.randint(len(nodes)) 
            if not nodes[index].is_given(): 
                node = nodes[index]
                break 
        if node  is None: 
            print_important("RandomNodesStrategy could not find a node after %d attempts, are you sure that the model has unobserved variables?"%(MAX_RANDOM_NODES))
        return nodes[index]





# Sampling methods
class RandomNumberGenerator():
    """Random number generator class. It includes all the functions used by the samplers for generation of random numbers. """
    def __init__(self): 
        pass    
    def rand_normal(self,mu,std):
        return numpy.random.normal(mu,std)
    def rand_multivariate_normal(self,mu,cov):
        return numpy.random.multivariate_normal(mu,cov) 
    def rand_uniform(self):
        return numpy.random.rand()


class SamplingMethod(object): 
    def __init__(self): 
        self._requires_log_probability                  = False
        self._requires_log_probability_gradient         = False
        self._requires_log_probability_hessian          = False
        self._requires_log_probability_diagonal_hessian = False
        self._requires_own_sampler                      = False 
        self._is_optimizer                              = True 
        # multi-purpose random number generator: 
        self.random_number_generator = RandomNumberGenerator()

    def requires_log_probability(self): 
        return self._requires_log_probability 
        
    def requires_log_probability_gradient(self): 
        return self._requires_log_probability_gradient 
        
    def requires_log_probability_hessian(self): 
        return self._requires_log_probability_hessian 
        
    def requires_log_probability_diagonal_hessian(self): 
        return self._requires_log_probability_diagonal_hessian 

    def requires_own_sampler(self): 
        return self._requires_own_sampler 

    def is_optimizer(self): 
        return self._is_optimizer 

    def can_be_used_with_node(self,node): 
        if not isinstance(node,Node): 
            raise UnexpectedParameterType("node must be an instance of Node.") 
        if self.requires_own_sampler(): 
            if not node.can_sample_conditional_probability(): 
                return False 
        if self.requires_log_probability(): 
            if not node.has_log_conditional_probability(): 
                return False
        if self.requires_log_probability_gradient(): 
            if not node.has_log_conditional_probability_gradient(): 
                return False 
        if self.requires_log_probability_hessian(): 
            if not node.has_log_conditional_probability_hessian(): 
                return False   
        if self.requires_log_probability_diagonal_hessian(): 
            if not node.has_log_conditional_probability_diagonal_hessian(): 
                return False    
        return True

    def get_name(self): 
        if hasattr(self,"__name__"): 
            return self.__name__
        elif hasattr(self,"__class__"): 
            return self.__class__.__name__ 
        else: 
            raise Exception("This should never ever happen. ") 

    # Subclass this function:
    def sample(self,graph,node,parameters=None): 
        #FIXME: implement
        print_debug("sample() not defined. ")
        return None

    # Subclass this function:
    def default_parameters(self): 
        return {}




class AncestralSampling(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(AncestralSampling, self).__init__(*args, **kwds) 
        self._requires_own_sampler = True
        self._is_optimizer = False

    def sample(self,graph,node,parameters=None): 
        print_debug("Sampling node '%s' with %s .."%(name(node),name(self)))
        sample = graph.sample_conditional_probability_node(node) 
        return sample
        
    def default_parameters(self): 
        return {}



class GradientAscent(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(GradientAscent, self).__init__(*args, **kwds) 
        self._requires_log_probability          = True
        self._requires_log_probability_gradient = True
        self._is_optimizer = True 

    def sample(self,graph,node,parameters=None):  
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sample = graph.get_node_value(node) 
        gradient = graph.get_log_conditional_probability_gradient(node,sample) 
        sample   = sample + parameters['step_size'] * gradient
        return sample

    def default_parameters(self): 
        return {'step_size':0.1}



class ExpectationMaximization(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(ExpectationMaximization, self).__init__(*args, **kwds) 
        self._requires_log_probability          = True 
        self._requires_log_probability_gradient = True 
        self._is_optimizer = True 

    def sample(self,graph,node,parameters=None):  
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sample = graph.get_node_value(node) 
        gradient = graph.get_log_conditional_probability_gradient(node,sample) 
        sample   = sample + gradient
        return sample

    def default_parameters(self): 
        return {} 



class QuasiNewton_L_BFGS_B(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(QuasiNewton_L_BFGS_B, self).__init__(*args, **kwds) 
        self._requires_log_probability = True
        self._requires_log_probability_gradient = True
        self._is_optimizer = True

    def sample(self,graph,node,parameters=None):
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()

        approximation_terms = parameters['approximation_terms'] 
        factr               = parameters['factr'] 
        pgtol               = parameters['pgtol'] 
        maxfun              = parameters['maxfun'] 
        self._graph         = graph
        self._node          = node 
        bounds              = parameters['bounds'] 
        sample              = numpy.float64( graph.get_node_value(node) )
        
        if bounds  is None:
            bounds = [(None,None)] * sample.size
        elif bounds == 'nonnegative':
            bounds = [(0,None)] * sample.size  
        else: 
            raise ParameterError("'bounds' parameter for QuasiNewton_L_BFGS_B.") 

        sample,f,d = fmin_l_bfgs_b(self._func, sample, fprime=self._grad, args=(), bounds=bounds, m=approximation_terms, factr=factr, pgtol=pgtol, maxfun=maxfun)
        print_debug("L_BFGS_B: %s (%d function calls)."%(d['task'],d['funcalls']))

        sample = numpy.asarray(sample)
        return sample

    def _func(self,sample,*args): 
        sample = numpy.asarray(sample)
        self._graph.set_node_value(self._node,sample) #FIXME: this line must be removed, but then there is a problem with l_bfgs_b
        return -numpy.float64(self._graph.get_log_conditional_probability(self._node)) 

    def _grad(self,sample,*args): 
        sample = numpy.asarray(sample)
        self._graph.set_node_value(self._node,sample) #FIXME: this line must be removed, but then there is a problem with l_bfgs_b
        return -numpy.float64(self._graph.get_log_conditional_probability_gradient(self._node)) 

    def default_parameters(self): 
        return {'approximation_terms':10, 'factr':1e6, 'pgtol':1e-8, 'maxfun':15000, 'bounds':None }



class Newton(SamplingMethod): 
    """Newton method, optimisation"""
    def __init__(self, *args, **kwds):
        super(Newton, self).__init__(*args, **kwds) 
        self._requires_log_probability          = True
        self._requires_log_probability_gradient = True
        self._requires_log_probability_hessian  = True
        self._is_optimizer = True 

    def sample(self,graph,node,parameters=None):  
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sample = graph.get_node_value(node)
        gradient = graph.get_log_conditional_probability_gradient(node,sample) 
        hessian  = graph.get_log_conditional_probability_hessian(node,sample) 
        sample   = sample + parameters['step_size'] * numpy.dot(numpy.linalg.inv(hessian),gradient.T).T 
        return sample

    def default_parameters(self): 
        return {'step_size':0.99}
                



class MetropolisHastingsMCMC(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(MetropolisHastingsMCMC, self).__init__(*args, **kwds) 
        self._requires_log_probability = True
        self._is_optimizer = False 

    def sample(self,graph,node,parameters=None): 
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sigma        = parameters['sigma']
        sample  = graph.get_node_value(node)
        
        candidate    = sample + self.random_number_generator.rand_normal(0,sigma*numpy.ones(sample.shape)) 
        alpha        = min(1,numpy.exp(graph.get_log_conditional_probability(node,candidate) - graph.get_log_conditional_probability(node,sample) ))
        u = self.random_number_generator.rand_uniform()
        if u < alpha:
            accept = True
            sample = candidate 
        else: 
            accept = False
            #sample = last_sample
        return sample

    def default_parameters(self): 
        return {'sigma':1}

           
class HamiltonianMCMC(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(HamiltonianMCMC, self).__init__(*args, **kwds) 
        self._requires_log_probability = True
        self._is_optimizer = False 

    def sample(self,graph,node,parameters=None): 
        # todo: implement!!
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sigma        = parameters['sigma']
        sample  = graph.get_node_value(node)
        
        candidate    = sample + self.random_number_generator.rand_normal(0,sigma*numpy.ones(sample.shape)) 
        alpha        = min(1,numpy.exp(graph.get_log_conditional_probability(node,candidate) - graph.get_log_conditional_probability(node,sample) ))
        u = self.random_number_generator.rand_uniform()
        if u < alpha:
            accept = True
            sample = candidate 
        else: 
            accept = False
            #sample = last_sample
        return sample

    def default_parameters(self): 
        return {'sigma':1}
 

class LangevinAdjustedMetropolisHastingsMCMC(SamplingMethod): 
    def __init__(self, *args, **kwds):
        super(LangevinAdjustedMetropolisHastingsMCMC, self).__init__(*args, **kwds) 
        self._requires_log_probability = True 
        #self._requires_log_probability_gradient = True
        #self._requires_log_probability_hessian = True
        self._is_optimizer = False 

    def sample(self,graph,node,parameters=None): 
        # todo: implement!!
        print_debug("Sampling node '%s' with %s ..\n"%(name(node),name(self)))
        if parameters  is None: 
            parameters = self.default_parameters()
        sigma        = parameters['sigma']
        sample  = graph.get_node_value(node)
        
        candidate    = sample + self.random_number_generator.rand_normal(0,sigma*numpy.ones(sample.shape)) 
        alpha        = min(1,numpy.exp(graph.get_log_conditional_probability(node,candidate) - graph.get_log_conditional_probability(node,sample) ))
        u = self.random_number_generator.rand_uniform()
        if u < alpha:
            accept = True
            sample = candidate 
        else: 
            accept = False
            #sample = last_sample
        return sample

    def default_parameters(self): 
        return {'sigma':1}
        




def best_sampling_method(graph,node,sampling_methods): 
    """Select the best sampling method out of a list of sampling methods. """
    # Use the preference expressed by the node and the properties of the node (gradient, ..) and of the sampling methods. 
    # 1) if available, use the node-specific preference:
    # FIXME: implement node preference 
    # 2) choose according to absolute preference criterium:   
    #   1- prefer a method based on direct sampling if node has a direct sampling method
    sampling_method = None
    if node.can_sample_conditional_probability(): 
        for method in sampling_methods: 
            if method.requires_own_sampler():
                sampling_method == method 
    #   2- prefer posterior sampling to optimisation: 
    #   3- prefer second order sampling strategies to first order (gradient based)
    #   4- prefer first order to order zero (likelihood based): 
    optimisers   = []
    samplers     = []
    if sampling_method  is None: 
        for method in sampling_methods: 
            if method.is_optimizer(): 
                optimisers.append(method)
            else:
                samplers.append(method)
        if node.has_log_conditional_probability_hessian(): 
            for method in samplers: 
                if method.requires_log_probability_hessian(): 
                    sampling_method = method 
        if sampling_method  is None: 
            if node.has_log_conditional_probability_diagonal_hessian(): 
                for method in samplers: 
                    if method.requires_log_probability_diagonal_hessian(): 
                        sampling_method = method             
            if sampling_method  is None: 
                if node.has_log_conditional_probability_gradient(): 
                    for method in samplers: 
                        if method.requires_log_probability_gradient(): 
                            sampling_method = method  
                if sampling_method  is None: 
                    if node.has_log_conditional_probability(): 
                        for method in samplers: 
                            if method.requires_log_probability(): 
                                sampling_method = method  
                    if sampling_method  is None: 
                        if node.has_log_conditional_probability_hessian(): 
                            for method in optimisers: 
                                if method.requires_log_probability_hessian(): 
                                    sampling_method = method 
                        if sampling_method  is None: 
                            if node.has_log_conditional_probability_diagonal_hessian(): 
                                for method in optimisers: 
                                    if method.requires_log_probability_diagonal_hessian(): 
                                        sampling_method = method             
                            if sampling_method  is None: 
                                if node.has_log_conditional_probability_gradient(): 
                                    for method in optimisers: 
                                        if method.requires_log_probability_gradient(): 
                                            sampling_method = method  
                                if sampling_method  is None: 
                                    if node.has_log_conditional_probability(): 
                                        for method in optimisers: 
                                            if method.requires_log_probability(): 
                                                sampling_method = method  
    print_debug("Best method for node %s: %s"%(name(node),name(sampling_method)) )
    return sampling_method



# Sampler state
class SamplerStateNode(): 
    """SamplerStateNode: stores the information relative to the sampling of a node. """ 
    def __init__(self,node_name): 
        self.node_name = node_name
        self.reset_state() 

    def reset_state(self): 
        self.n_accepted       = 0 
        self.n_rejected       = 0
        self.n_samples        = 0 
        self.rejection_ratio  = 0.0 
        self.sample_ratio     = 0.0   # ratio between number of samples of the node and number of samples of the graph
        self.last_method        = None
        self.last_threshold     = None
        self.last_random_number = None
        self.last_seed          = None
        self.last_parameters    = None
        self.last_accepted      = None

    def to_dictionary(self): 
        dictionary =  {
                       'node_name':self.node_name,
                       'n_accepted':self._n_accepted, 
                       'n_rejected':self._n_rejected, 
                       'n_samples':self._n_samples, 
                       'rejection_ratio':self.rejection_ratio, 
                       'sample_ratio':self.sample_ratio, 
                       'last_threshold':self.last_threshold, 
                       'last_random_number':self.last_random_number, 
                       'last_seed':self.last_seed, 
                       'last_parameters':self.last_parameters, 
                       'last_accepted':self.last_accepted, 
                       'last_method':self.last_method } 
        return dictionary

    def to_pickle(self): 
        dictionary = self.to_dictionary() 
        return pickle.dumps(dictionary) 

    def new_sample(self,method,accepted,threshold,random_number,seed,parameters,nsamples_graph):      
        self.n_samples += 1
        if accepted: 
            self.n_accepted +=1
        else:
            self.n_rejected +=1 
        self.last_method = method
        self.last_threshold = threshold
        self.last_random_number = random_number 
        self.last_seed = seed 
        self.last_parameters = parameters
        self.last_accepted = bool(accepted) 
        self.rejection_ratio = self.n_rejected/(1.0*self.n_samples)
        self.sample_ratio = nsamples_graph/(1.0*self.n_samples) 

    def _repr_html_(self):
        table_data = [['node_name',self.node_name,"name of the node"], 
        ['last_method',self.last_method,"sampling method used to obtain the last sample"], 
        ['n_samples',self.n_samples,"total number of samples"], 
        ['n_accepted',self.n_accepted,"number of samples accepted"], 
        ['n_rejected',self.n_rejected,"number of samples rejected"], 
        ['rejection_ratio',self.rejection_ratio,"n_rejected / n_samples"], 
        ['last_accepted',self.last_accepted,"indicates if the last sample was accepted"], 
        ['sample_ratio',self.sample_ratio,"n_samples of the node / n_samples of the graph"], 
        ['last_threshold',self.last_threshold,"threshold value (Metropolis Hastings) for the last sample"], 
        ['last_random_number',self.last_random_number,"random number for the last sample"], 
        ['last_seed',self.last_seed,"seed for the last sample"], ]
        for param in self.last_parameters.keys(): 
            table_data += [ [param,self.last_parameters[param],"parameter of the last sampling method"] ] 
        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic_left')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        return table._repr_html_()
        

class SamplerState():
    """SamplerSate: stores the state of the Sampler, including information about 
    each node and global information. """
    def __init__(self): 
        self.reset_state() 

    def reset_state(self): 
        self.n_accepted = 0 
        self.n_rejected = 0
        self.n_samples  = 0 
        self.rejection_ratio = 0.0
        self.last_node     = None
        self.last_accepted = None
        self._nodes_state = {} 

    def _has_node(self,nodename):
        return self._nodes_state.has_key(nodename) 
    
    def get_n_nodes(self): 
        return len(self._nodes_state) 

    def get_state_node(self,nodename): 
        if not self._has_node(nodename): 
            raise ParameterError("There is no information regarding node '%s'"%nodename)  
        return self._nodes_state[nodename]  
    
    def new_sample(self,nodename,method,accepted,threshold,random_number,seed,parameters): 
        # if new node name was previously unknown, create record for the node 
        if not self._has_node(nodename): 
            self._nodes_state[nodename] = SamplerStateNode(nodename) 
            
        # update node state
        node_state = self._nodes_state[nodename]
        node_state.new_sample(method,accepted,threshold,random_number,seed,parameters,self.n_samples) 
        self._nodes_state[nodename] = node_state 
        
        # update global sampler state
        self.n_samples += 1
        if accepted: 
            self.n_accepted +=1
        else:
            self.n_rejected +=1 
        self.last_node     = nodename 
        self.last_accepted = bool(accepted) 
        self.rejection_ratio = self.n_rejected/(1.0*self.n_samples)        

    def to_dictionary(self): 
        dictionary =  {
                       'n_nodes':self.get_n_nodes(),
                       'n_accepted':self._n_accepted, 
                       'n_rejected':self._n_rejected, 
                       'n_samples':self._n_samples, 
                       'rejection_ratio':self.rejection_ratio, 
                       'last_accepted':self.last_accepted, 
                       'last_method':self.last_method } 
        return dictionary

    def to_pickle(self): 
        dictionary = self.to_dictionary() 
        return pickle.dumps(dictionary) 

    def _repr_html_(self):
        table_data = [['n_nodes',self.get_n_nodes(),"number of nodes recorded"], 
        ['n_samples',self.n_samples,"total number of samples"], 
        ['n_accepted',self.n_accepted,"number of samples accepted"], 
        ['n_rejected',self.n_rejected,"number of samples rejected"], 
        ['rejection_ratio',self.rejection_ratio,"n_rejected / n_samples"], 
        ['last_accepted',self.last_accepted,"indicates if the last sample was accepted"], ]
        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic_left')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        return table._repr_html_()

    
             
# Sampler 
class Sampler(object):
    """Sampler: samples from a ProbabilisticGraphicalModel. """
    def __init__(self,graph,optimization_only=False): 
        self._nodes_compatible_sampling_methods = {}
        self._nodes_active_sampling_method      = {}
        self._nodes_sampling_parameters         = {}
        self._sampling_methods = [] 
        self.__load_sampling_methods() 
        self.graph = None
        if graph  is not None: 
            self.attach_to_graph(graph,optimization_only)
        # By default, alternate nodes randomly:  
        self.set_sampling_strategy(RandomNodesStrategy)
        # By default, trace in RAM 
        self.attach_to_tracer( RamTracer() )
        # Instantiate state:
        self.state = SamplerState() 

    def _has_graph(self): 
        return isinstance(self.graph,ProbabilisticGraphicalModel) 
 
    def attach_to_graph(self,graph,optimization_only=False): 
        if not isinstance(graph,ProbabilisticGraphicalModel):
            raise UnexpectedParameterType("graph must be an instance of ProbabilisticGraphicalModel. ") 
        self.graph = graph 
        # for each node, determine which sampling methods can be used  
        for node in self.graph.get_nodes(): 
            sampling_methods = []
            for sampling_method in self._sampling_methods: 
                if sampling_method.can_be_used_with_node(node): 
                    if not optimization_only: 
                        sampling_methods.append(sampling_method) 
                    else: 
                        if sampling_method.is_optimizer(): 
                            sampling_methods.append(sampling_method) 
            self._nodes_compatible_sampling_methods[name(node)] = sampling_methods
            if sampling_methods == []: 
                print_important('None of the sampling methods can sample from node %s.\n'%name(node) ) 
        # by default set the sampling method automatically for all nodes
        self.set_sampling_method_auto(optimization_only) 
        self.__define_dynamic_methods() 

    def attach_to_tracer(self,tracer): 
#        if not isinstance(tracer,Tracer): 
#            raise UnexpectedParameterType("tracer must be an instance of Tracer. ") 
        self.tracer = tracer 
         
    def set_node_sampling_method_auto(self,node,optimization_only=False):  
        if not self.graph.has_node(node): 
            raise ParameterError("The graph does not contain node '%s'."%name(node)) 
        # filter the methods if only optimization is requested
        if optimization_only: 
            sampling_methods = []
            for sampling_method in self._nodes_compatible_sampling_methods[name(node)]: 
                if sampling_method.is_optimization(): 
                    sampling_mehtods.append(sampling_methods)
        else: 
            sampling_methods = self._nodes_compatible_sampling_methods[name(node)]
#        print_debug("The samplers compatible with are: %s\n"%name(self._nodes_compatible_sampling_methods[name(node)]) )
        # choose the best sampler 
        self._nodes_active_sampling_method[name(node)] = best_sampling_method(self.graph,node,sampling_methods)  
        # use the default parameters for the sampling method 
        self._nodes_sampling_parameters[name(node)] = self._nodes_active_sampling_method[name(node)].default_parameters()
#        print_debug("Selected sampling method '%s' for node '%s'. \n"%( name(self._nodes_active_sampling_method[name(node)]),name(node)) )
        return self._nodes_active_sampling_method[name(node)] 

    def set_sampling_method_auto(self,optimization_only=False): 
        for node in self.graph.get_nodes(): 
            self.set_node_sampling_method_auto(node,optimization_only) 

    def get_sampling_method_node(self,node): 
        return self._nodes_active_sampling_method[name(node)]

    def get_sampling_parameters_node(self,node): 
        return self._nodes_sampling_parameters[name(node)]

    def set_node_sampling_method_manual(self,node,sampling_method_name,parameters=None): 
        if not self.graph.has_node(node): 
            raise ParameterError("The graph does not contain node '%s'."%name(node)) 
        if not sampling_method_name in map(name,self._nodes_compatible_sampling_methods[name(node)]): 
            raise ParameterError("Sampling method '%s' is not compatible with node '%s'."%(sampling_method_name,name(node)))
        self._nodes_active_sampling_method[name(node)] = self._nodes_compatible_sampling_methods[name(node)][map(name,self._nodes_compatible_sampling_methods[name(node)]).index(sampling_method_name)] 
        if parameters  is None: 
            parameters = self._nodes_active_sampling_method[name(node)].default_parameters()
        self._nodes_sampling_parameters[name(node)] = parameters 

    def sample_node(self, node, nsamples=1, trace=True, parameters=None, display_progress=True): 
        node = name(node) 
        # check if the node exists: 
        if not self.graph.has_node(node): 
            raise ParameterError("The graph does not contain node '%s'."%node) 
        # if node has given value, just return the value: 
        node_is_given = self.graph.get_node(node).is_given()
        # check if a sampling strategy for the given node has been selected: 
        if not self._nodes_active_sampling_method.has_key(node): 
            raise NotInitialized("The sampling strategy for node '%s' has not beel selected. "%node) 
        # initialize progress bar: 
        if display_progress: 
            progress_bar = ProgressBar(height='6px', width='100%%', background_color=LIGHT_RED, foreground_color=RED)
        # sample: 
        if not node_is_given: 
            active_sampling_method = self._nodes_active_sampling_method[node] 
            for i in range(nsamples):
                # update progress bar
                if display_progress: 
                    progress_bar.set_percentage(i*100.0/nsamples)
                sample = active_sampling_method.sample(self.graph,node,parameters) 
                # update state of the graphical model: 
                self.graph.set_node_value(node,sample) 
                # update sampler state
                self.state.new_sample(name(node),name(active_sampling_method),True,0.2,2.24,234,{'delta':0.01}) 
                #FIXME: fix line above: make the samplers return that information 
                # trace
                if trace: 
                    self.tracer.new_sample(name(node),sample )
        else: 
            sample = self.graph.get_node_value(node)
            if trace:
                for i in range(nsamples):
                    # update state of the graphical model: 
                    self.graph.set_node_value(node,sample) 
                    # update sampler state 
                    self.state.new_sample(name(node),"given",True,0.2,2.24,234,{'delta':0.01}) 
                    # trace 
                    self.tracer.new_sample(name(node),sample )      
                    # update progress bar
                    if display_progress: 
                        progress_bar.set_percentage(i*100.0/nsamples)
        if display_progress: 
            progress_bar.set_percentage(100.0)                     
        return sample 

    def sample(self,nsamples=1,trace=True,nsamples_per_node=1,display_progress=True): 
        if display_progress: 
            progress_bar = ProgressBar(height='6px', width='100%%', background_color=LIGHT_RED, foreground_color=RED)
        # alternate the nodes according to the sampling strategy: 
        for i in range(nsamples): 
            if display_progress: 
                progress_bar.set_percentage(i*100.0/nsamples)
            node = self._sampling_strategy.next_node() 
            sample = self.sample_node(node,nsamples_per_node,trace,display_progress=False) 
        if display_progress: 
            progress_bar.set_percentage(100.0)
        return sample 

    def set_sampling_strategy(self,sampling_strategy): 
        if not issubclass(sampling_strategy,SamplingStrategy): 
            raise UnexpectedParameterType("sampling_strategy must be a class of type SamplingStrategy. ")
        self._sampling_strategy = sampling_strategy(self.graph) 

    def get_sampling_strategy(self): 
        self._sampling_strategy 

    def get_state_node(self,nodename): 
        return self.state.get_state_node(nodename) 
 
    def __define_dynamic_methods(self):
        nodenames = self.graph.list_nodes() 
        for nodename in nodenames: 
            def func(nsamples=1, trace=True, parameters=None, nodename=nodename): 
                return self.sample_node(nodename,nsamples,trace,parameters)
            setattr(self,"sample_"+nodename,func)
        
    def __load_sampling_methods(self): 
        self._sampling_methods = []   
        for name,obj in inspect.getmembers(sys.modules[__name__], inspect.isclass):
           if issubclass(obj,SamplingMethod): 
               if not obj == SamplingMethod: 
                   self._sampling_methods.append(obj()) 

    def _list_sapling_methods_names(self): 
        l = []
        for m in self._sampling_methods: 
            l.append(name(m))
        return l 

    sampling_methods = property(_list_sapling_methods_names, None, None, "Empty doc. ") 






# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# 20 Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging
# Dec 2013, Boston, MA, USA 


#__all__ = ['Graphs','Samplers','Tracers','Display','Models','Sources','examples']
#from ilang import Graphs, Samplers, Tracers, Display, Models, Sources
import examples





# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


__all__ = ['RamTracer']

from ilang.Display import DisplayPylab
from ilang.exceptions import *
from ilang.verbose import * 

import numpy


class Tracer(): 
    def __init__(self): 
        self.reset_trace() 

    def reset_trace(self):
        pass 


class RamTracer(Tracer): 
    """Stores the trace of a sampler in RAM memory. """
    def __init__(self,batch_size=1000): 
        self._batch_size  = batch_size # allocate new memory evey '_batch_size' samples 
        self.reset_trace() 

    def reset_trace(self): 
        self._samples     = None       # samples 
        self._shape       = None       # shape of the sample array
        self._index       = 0          # index of the last sample 
        self._threshold        = None 
        self._accepted         = None 
        self._acceptance_ratio = None
        self._batch_index = 0
        self._size        = 0          # size of the sample 
        # Default display: 
        self.display      = DisplayPylab(self) 

    def get_trace(self):
        return self

    def empty(self): 
        if self._samples  is None or self._shape  is None or self._index == 0: 
            return True
        else: 
            return False

    def shape(self):
        if self.empty(): 
            return (0,)
        else:
            return self._shape

    def size(self):
        if self.empty(): 
            return (0,)
        else:
            return self._size
        
    def change_batch_size(self,batch_size): 
        self._batch_size = batch_size

    def set_shape(self,shape):
        if not numpy.asarray(shape).prod() == self._size: 
            print "Not a valid shape, not compatible with sample size" 
            return False 
        else: 
            self._shape = shape 
            return True
    def scale(self,factor): 
        for i in range(self._index): 
            self._samples[i] = self._samples[i]*factor

    def new_sample(self,node_name,sample,parameters=None): 
        if self._index == 0: 
            # first sample, set the sample size and shape 
            self._size  = sample.size 
            self._shape = sample.shape
            self._samples = numpy.zeros((self._batch_size,self._size)) 
            self._threshold        = numpy.zeros(self._batch_size)  
            self._accepted         = numpy.zeros(self._batch_size)
            self._acceptance_ratio = numpy.zeros(self._batch_size)
        else: 
            if self._batch_index == 0: 
                # allocate new batch
                self._samples = numpy.vstack( (self._samples,numpy.zeros((self._batch_size,self._size))) )
#                self._threshold = numpy.hstack( (self._threshold,numpy.zeros(self._batch_size)) )
#                self._accepted = numpy.hstack( (self._accepted,numpy.zeros(self._batch_size)) )
#                self._acceptance_ratio = numpy.hstack( (self._acceptance_ratio,numpy.zeros(self._batch_size)) )
#        print "Size: ",sample.size, sample.shape, self._size, self._shape
        self._samples[self._index,:] = sample.reshape((self._size,))
#        self._threshold[self._index] = threshold
#        self._accepted[self._index] = accepted
#        self._acceptance_ratio[self._index] = acceptance_ratio
        self._index += 1
        self._batch_index += 1 
        if self._batch_index == self._batch_size: 
            self._batch_index = 0 

    def get_samples(self): 
        if self.empty(): 
            return None
        return self._samples[0:self._index,:].reshape((self._index,)+self._shape)

    def get_last_sample(self): 
        return self._samples[self._index-1,:].reshape(self._shape)

    def get_threshold(self): 
        if self.empty(): 
            return None
        return self._threshold[0:self._index]

    def get_accepted(self): 
        if self.empty(): 
            return None
        return self._accepted[0:self._index]

    def get_acceptance_ratio(self): 
        if self.empty(): 
            return None
        return self._acceptance_ratio[0:self._index]

    def n_samples(self): 
        if self.empty(): 
            return 0
        return self._index

    def min(self): 
        if self.empty(): 
            return None
        return self._samples[0:self._index,:].min()

    def max(self):
        if self.empty(): 
            return None
        return self._samples[0:self._index,:].max()

    def mean(self): 
        if self.empty(): 
            return None
        return numpy.mean(self._samples[0:self._index,:],0).reshape(self._shape) 

    def mean_at_location(self,coordinate): 
        if self.empty(): 
            return None
        if isinstance((coordinate),type(1)): 
            coordinate = (coordinate,)
        if len(coordinate)>1:
            coordinate = numpy.ravel_multi_index(coordinate,self._shape) 
        return numpy.mean( self.get_samples().reshape((self._index,self._size))[:,coordinate])   

    def variance(self): 
        if self.empty(): 
            return None
        return numpy.var(self._samples[0:self._index,:],0).reshape(self._shape) 

    def covariance(self): 
        if self.empty(): 
            return None
        return numpy.cov(self._samples[0:self._index,:].T)

    def covariance_line(self,coordinates,covariance=None): 
        if self.empty(): 
            return None
        if not covariance: 
            covariance = self.covariance()    
        if isinstance((coordinate),type(1)): 
            coordinate = (coordinate,)
        if len(coordinates)>1:         
            coordinate = numpy.ravel_multi_index(coordinates,self._shape)
        return covariance[coordinate,:].reshape(self._shape)

    def histogram_bivariate(self,coordinate_A,coordinate_B,n_bins=(100,100),range_histogram=None):       
        if self.empty(): 
            return None
        if range_histogram  is None: 
            range_histogram = [(0,1.2*self.max()),(0,1.2*self.max())] 
        if isinstance((coordinate_A),type(1)): 
            coordinate_A = (coordinate_A,)
        if len(coordinate_A)>1:
            coordinate_A = numpy.ravel_multi_index(coordinate_A,self._shape) 
        if isinstance((coordinate_B),type(1)): 
            coordinate_B = (coordinate_B,)
        if len(coordinate_B)>1:
            coordinate_B = numpy.ravel_multi_index(coordinate_B,self._shape) 
        return numpy.histogram2d((self.get_samples().reshape((self._index,self._size))[:,coordinate_A]).reshape(self._index), (self.get_samples().reshape((self._index,self._size))[:,coordinate_B]).reshape(self._index),bins=n_bins,range=range_histogram)        

    def histogram(self,coordinate,n_bins=100,range_histogram=None): 
        if self.empty(): 
            return None
        if range_histogram  is None: 
            range_histogram = (0,1.2*self.max())
        if isinstance((coordinate),type(1)): 
            coordinate = (coordinate,)
        if len(coordinate)>1:
            coordinate = numpy.ravel_multi_index(coordinate,self._shape) 
        return numpy.histogram(self.get_samples().reshape((self._index,self._size))[:,coordinate],bins=n_bins,range=range_histogram)   

    def histogram_roi_integral(self,roi,n_bins=100,range_histogram=None): 
        if self.empty(): 
            return None
        roi_integral_samples= self.roi_integral_samples(roi) 
        if range_histogram  is None: 
            range_histogram = (0,1.1*roi_integral_samples.max())
        return numpy.histogram(roi_integral_samples,bins=n_bins,range=range_histogram)   

    def roi_samples(self,roi): 
        x0,y0 = roi[0]
        x1,y1 = roi[1]
        return self.get_samples()[:,x0:x1,y0:y1]

    def roi_integral_samples(self,roi): 
        return self.roi_samples(roi).sum(1).sum(1) 

    def extract_roi_from_sample(self,roi,sample): 
        x0,y0 = roi[0]
        x1,y1 = roi[1]
        return sample[x0:x1,y0:y1]

    def mean_roi(self,roi): 
        if self.empty(): 
            return None
        roi_samples = self.roi_samples(roi)
        return numpy.mean(roi_samples[0:self._index,:],0).reshape(roi_samples[0].shape) 

    def variance_roi(self,roi): 
        if self.empty(): 
            return None
        roi_samples = self.roi_samples(roi)
        return numpy.var(roi_samples[0:self._index,:],0).reshape(roi_samples[0].shape) 

    def autocorrelation(self,coordinate): 
        if self.empty(): 
            return None
        if isinstance((coordinate),type(1)): 
            coordinate = (coordinate,)
        if len(coordinate)>1:
            coordinate = numpy.ravel_multi_index(coordinate,self._shape) 
        samples = self.get_samples().reshape((self._index,self._size))[:,coordinate] 
        samples = (samples-numpy.mean(samples)).reshape(samples.size)
        autocorr_f = numpy.correlate(samples, samples, mode='full')
        return autocorr_f[autocorr_f.size/2:]/autocorr_f[autocorr_f.size/2] 



# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


import unittest
from .. import Graphs
from .. import exceptions
from .. import verbose

verbose.set_verbose_low()


class TestSequenceProbabilisticGraphicalModel(unittest.TestCase):
    """Sequence of tests for ProbabilisticGraphicalModel class"""
    
    def setUp(self):
        pass
    
    def test_ProbabilisticGraphicalModel_nodes(self):
        """Test whether ProbabilisticGraphicalModel correctly manages inserting and deleting nodes"""
        dag = Graphs.ProbabilisticGraphicalModel() 
        # 1) add node instance 
        node1 = Graphs.Node('node1',value=0,given=False); 
        dag.add_nodes(node1) 
        # 2) add node by name
        self.assertEqual(node1.name,dag.get_node('node1').get_name())
        dag.add_nodes('node2')
        node2 = dag.get_node('node2') 
        self.assertTrue(node2 in dag.get_nodes())         
        # 3) check if exception is thrown in case of duplicate node names 
        self.assertRaises(exceptions.ParameterError, dag.add_nodes, 'node1')
        self.assertRaises(exceptions.ParameterError, dag.add_nodes, node2) 
        # 4) remove node 
        self.assertRaises(exceptions.ParameterError, dag.remove_node, 'node3')
        dag.remove_node('node2')
        self.assertEqual(dag.get_nodes(),[node1,]) 
        # 5) add multiple nodes 
        dag.add_nodes(['node3','node4'])
        self.assertTrue('node3' in dag.get_nodes_names() and 'node4' in dag.get_nodes_names()) 
   
    def test_ProbabilisticGraphicalModel_dependencies(self): 
        """Test whether ProbabilisticGraphicalModel correctly manages attaching and detaching dependencies"""
        dag = Graphs.ProbabilisticGraphicalModel() 
        # 1) attach a dependence 
        
        # 2) detach a dependence 
        
        # 3) connect a dependence 
        
        # 4) fully connected property 
        
        # 5) complete graph property 
        
    def test_ProbabilisticGraphicalModel_inference(self): 
        """Test methods of ProbabilisticGraphicalModel related to inference"""
        # 1) test extraction of Markov blanket
        pass
        # 2) test d-separation property 
        
        # 3) test parents, children 
        
        
    def test_ProbabilisticGraphicalModel_export(self): 
        """Test whether ProbabilisticGraphicalModel correctly manages exporting graph"""
        # 1) export dictionary
        pass


        
        
if __name__ == '__main__':
    unittest.main()







# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


from ilang.Graphs import Dependence, ProbabilisticGraphicalModel
from ilang.Models import Poisson, Smoothness 
from ilang.Samplers import Sampler
from ilang.Tracers import RamTracer
from ilang.Display import MatplotlibDisplay
import numpy



# Define the model components 
observation = Poisson('SPECT')
prior_activity = Smoothness('Smoothing_Activity') 
prior_attenuation = Smoothness('Smoothing_Attenuation')  

# Build the graph 
dag = ProbabilisticGraphicalModel(['activity','attenuation','counts','smoothing-activity','smoothing-attenuation']) 
dag.set_nodes_given(['counts','smoothing-activity','smoothing-attenuation'], True)
dag.add_dependence(observation,{'lambda':'activity','alpha':'attenuation','z':'counts'})
dag.add_dependence(prior_activity,{'x':'activity','beta':'smoothing-activity'}) 
dag.add_dependence(prior_attenuation,{'x':'attenuation','beta':'smoothing-attenuation'}) 

# Initialize the nodes of the graph   
dag.set_node_value('activity',numpy.ones((10,10)))

# Instantiate the sampler, tracer and display 
sampler = Sampler(dag)
tracer = RamTracer(sampler)  
display = MatplotlibDisplay(tracer) 

# Sample 
sampler.sample(1000,trace=False) 
display.imagesc_node('activity')



if __name__=="__main__": 
    dag.webdisplay(background=False) 






# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


# Exceptions 
class UnexpectedParameterType(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr("Unexpected parameter type: "+str(self.value))
        
class ParameterError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr("Parameter error: "+str(self.value))

class InconsistentGraph(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr("Parameter error: "+str(self.value))        
        
class NotInitialized(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr("Not initialized: "+str(self.value)) 

class NoCompatibleSampler(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr("Not compatible sampler: "+str(self.value)) 

class ModelUndefined(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return repr("Model method undefined: "+self.msg) 



# iLang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Martinos Center for Biomedical Imaging - MGH/Harvard University, Jan 2015


from setuptools import setup
from glob import glob 

setup(
    name='ilang',
    version='0.3.1',
    author='Stefano Pedemonte',
    author_email='stefano.pedemonte@gmail.com',
	packages=['ilang', 'ilang.test','ilang.examples',],
    test_suite="ilang.test",
    scripts=[],
    url='http://www.occiput.io',
    license='LICENSE.txt',
    description='Inference Language for Volumetric Imaging',
    long_description=open('README.rst').read(),
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=[
        "numpy >= 1.6.0",
        #"pil >= 1.0.0",
        "DisplayNode >= 0.3.0", 
        "nibabel >= 2.0.0",
        "ipy_table >= 1.11.0", 
    ],
)






# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


import unittest
from .. import Tracers
from .. import exceptions
from .. import verbose

verbose.set_verbose_low()

class TestSequenceTracer(unittest.TestCase):
    """Sequence of tests for the Tracer sampler"""
    def setUp(self):
        pass
    
    def test_trace(self):
        """.."""
        pass




if __name__ == '__main__':
    unittest.main()







# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


# Print with 3 levels of verbosity 
verbose = 1
def set_verbose_high(): 
    """Print everything - DEBUG mode"""
    global verbose; verbose = 2    
def set_verbose_medium(): 
    """Print runtime information"""
    global verbose; verbose = 1
def set_verbose_low(): 
    """Print only important messages"""
    global verbose; verbose = 0
def set_verbose_no_printing(): 
    """Do not print messages at all"""
    global verbose; verbose = -1        
def get_verbose_level():
    return verbose
def print_debug(msg):
    """Use this for DEBUG Information"""
    if verbose >= 2: 
        print msg
def print_runtime(msg):
    """Use this for messages useful at runtime"""
    if verbose >= 1: 
        print msg
def print_important(msg):
    """Use this for important messages"""
    if verbose >= 0: 
        print msg



# ilang - Inference Language 
# Stefano Pedemonte
# Oct. 2013, Alessandria, Italy



__all__ = ['DisplayPylab'] 

from ilang.exceptions import *
from ilang.verbose import * 

try:
    import pylab 
except:
    print "Please install Pylab"
    has_pylab = False
else: 
    has_pylab = True
    #pylab.ion()
    from colorsys import hsv_to_rgb 
try:
    import matplotlib.pyplot as plt
    plt.ion()
except: 
    pass



class DisplayPylab(): 
    """Display information about the samples stored in a tracer. """
    def __init__(self,tracer): 
        self._tracer = tracer


    def plot_mean(self,figure=None,cmap=None,title="Mean",log_scale=False,interpolation='none'): 
        mean = self._tracer.mean() 
        if mean  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                mean = np.log(mean-mean.min()+1e-10)
            pylab.imshow(mean,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_variance(self,figure=None,cmap=None,title="Variance",log_scale=False,interpolation='none'): 
        variance = self._tracer.variance() 
        if variance  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                variance = np.log(variance-variance.min()+1e-10)
            pylab.imshow(variance,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_sample_index(self,index,figure=None,cmap=None,title=None,log_scale=False,interpolation='none'): 
        samples = self._tracer.get_samples()
        if samples is None: 
            return False
        sample = samples[index] 
        if title is None: 
            title = "Sample %d"%index 
        return self.plot_sample_value(sample,figure,cmap,title,log_scale,interpolation) 

    def plot_sample_value(self,sample,figure=None,cmap=None,title="Sample",log_scale=False,interpolation='none'): 
        if not figure:
            pass #figure = pylab.figure()
        else: 
            pylab.figure(figure.number)
        if cmap  is None: 
            cmap = pylab.cm.Greys_r
        pylab.imshow(sample,cmap,interpolation=interpolation)
        pylab.title(title)
        #pylab.draw()
        return figure

    def plot_mean_roi(self,roi,figure=None,cmap=None,title="Mean ROI",log_scale=False,interpolation='none'): 
        mean = self._tracer.mean_roi(roi) 
        if mean  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                mean = np.log(mean-mean.min()+1e-10)
            pylab.imshow(mean,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_variance_roi(self,roi,figure=None,cmap=None,title="Variance ROI",log_scale=False,interpolation='none'): 
        variance = self._tracer.variance_roi(roi) 
        if variance  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                variance = np.log(variance-variance.min()+1e-10)
            pylab.imshow(variance,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_covariance(self,figure=None,cmap=None,title="Covariance",log_scale=False,interpolation='none'): 
        covariance = self._tracer.covariance() 
        if covariance  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                covariance = np.log(covariance-covariance.min()+1e-10)
            pylab.imshow(covariance,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_covariance_line(self,xy,covariance=None,figure=None,cmap=None,title="Covariance",log_scale=False,interpolation='none'): 
        covariance_line = self._tracer.covariance_line(xy,covariance)             
        if covariance_line  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                covariance_line = np.log(covariance_line-covariance_line.min()+1e-10)
            pylab.imshow(covariance_line,cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_histogram_bivariate(self,coordinates_A,coordinates_B,n_bins=(100,100),range_plot=None,figure=None,cmap=None,title="Histogram",log_scale=False,interpolation='bicubic',subplot=None):       
        histogram,x,y = self._tracer.histogram_bivariate(coordinates_A,coordinates_B,n_bins,range_plot)             
        if histogram  is None: 
            return False
        else: 
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if subplot is not None: 
                pylab.subplot(*subplot)
            if cmap  is None: 
                cmap = pylab.cm.Greys_r
            if log_scale: 
                histogram = np.log(histogram+1e-10)
            pylab.imshow(histogram,cmap=cmap,interpolation=interpolation)
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_histogram(self,coordinates,n_bins=100,range_plot=None,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title="Histogram",true_value=None,estimated_value=None,subplot=None): 
        histogram,x = self._tracer.histogram(coordinates,n_bins,range_plot) 
        if histogram  is None: 
            return False
        else: 
            if color_line  is None: 
                color_line = hsv_to_rgb(0.0,0.85,1)
            if color_fill  is None: 
                color_fill = hsv_to_rgb(0.0,0.85,1)
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if subplot is not None: 
                pylab.subplot(*subplot)
            pylab.plot(x[0:-1],histogram,alpha=alpha_line,color=color_line)
            pylab.fill_between(x[0:-1], 0, histogram, alpha=alpha_fill, facecolor=color_fill) 

            mean_value = self._tracer.mean_at_location(coordinates)
            pylab.axvline(x=mean_value, ymin=0.0, ymax=0.99,alpha=alpha_line,color=color_line,linewidth=1,linestyle='dashed')

            if true_value is not None: 
                if not isinstance(true_value,type(1)): 
                    true_value = true_value[coordinates[0],coordinates[1]]
                pylab.axvline(x=true_value, ymin=0.0, ymax=0.99,alpha=alpha_line,color=color_line,linewidth=1,linestyle='solid')
            if estimated_value is not None: 
                if not isinstance(true_value,type(1)): 
                    estimated_value = estimated_value[coordinates[0],coordinates[1]]
                pylab.axvline(x=estimated_value, ymin=0.0, ymax=0.99,alpha=alpha_line,color=color_line,linewidth=1,linestyle='dotted')
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_histogram_roi_integral(self,roi,n_bins=100,range_plot=None,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title="ROI integral",true_value=None,estimated_value=None): 
        histogram,x = self._tracer.histogram_roi_integral(roi,n_bins,range_plot) 
        if histogram  is None: 
            return False
        else: 
            if color_line  is None: 
                color_line = hsv_to_rgb(0.0,0.85,1)
            if color_fill  is None: 
                color_fill = hsv_to_rgb(0.0,0.85,1)
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            pylab.plot(x[0:-1],histogram,alpha=alpha_line,color=color_line)
            pylab.fill_between(x[0:-1], 0, histogram, alpha=alpha_fill, facecolor=color_fill)
            figure.hold(1)
            if true_value is not None:
                if not isinstance(true_value,type(1)): 
                    true_value = self._tracer.extract_roi_from_sample(roi,true_value).sum()
                pylab.axvline(x=true_value, ymin=0.0, ymax=0.99,color=color_line,alpha=alpha_line,linewidth=1)
            if estimated_value is not None:
                if not isinstance(true_value,type(1)): 
                    estimated_value = self._tracer.extract_roi_from_sample(roi,estimated_value).sum()
                pylab.axvline(x=estimated_value, ymin=0.0, ymax=0.99,color=color_line,alpha=alpha_line,linewidth=1,linestyle='dashed')
            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_profile(self,axis,index,n_bins=100,range_plot=None,true_value=None,estimated_value=None,figure=None,cmap=None,color_line=None,title="Profile"): 
        #FIXME: implementation is 2-D only, extend to 3-D
        if axis!=0 and axis!=1: 
            raise UnexpectedParameter("axis: 0 or 1")
        if cmap is None:
            cdict = {'red': ((0.0, 1.0, 1.0),(1.0, 0.7, 0.7)),'green': ((0.0, 1.0, 1.0),(1.0, 0.0, 0.0)),'blue': ((0.0, 1.0, 1.0),(1.0, 0.0, 0.0))} 
            cmap = pylab.matplotlib.colors.LinearSegmentedColormap('my_colormap',cdict,256)
        shape = self._tracer.shape() 
        n_points = shape[1-axis]
        M=np.zeros((n_points,n_bins)) 
        if range_plot is None: 
            range_plot=(self._tracer.min(),self._tracer.max())
        for i in range(n_points): 
            if axis==0:
                M[i,:] = self._tracer.histogram((index,i),n_bins=100,range_histogram=range_plot)[0]
            else: 
                M[i,:] = self._tracer.histogram((i,index),n_bins=100,range_histogram=range_plot)[0]
        if axis==0:
            p = self._tracer.get_samples()[:,index,:]
            if true_value is not None:
                p_t = true_value[index,:] 
            if estimated_value is not None:
                p_e = estimated_value[index,:] 
        else: 
            p = self._tracer.get_samples()[:,:,index]
            if true_value is not None:
                p_t = true_value[:,index] 
            if estimated_value is not None:
                p_e = estimated_value[:,index] 
        p_m=np.mean(p,0) 
        if color_line  is None: 
            color_line = hsv_to_rgb(0.2,0.0,0.3)
        if not figure:
            pass #figure = pylab.figure()
        else: 
            pylab.figure(figure.number)
        pylab.hold(1)
        pylab.plot((n_bins/(range_plot[1]-range_plot[0]))*p_m,color=color_line,linewidth=1,linestyle='dashed')
        if true_value is not None:
            pylab.plot((n_bins/(range_plot[1]-range_plot[0]))*p_t,color=color_line,linewidth=1,linestyle='solid')
        if estimated_value is not None:
            pylab.plot((n_bins/(range_plot[1]-range_plot[0]))*p_e,color=color_line,linewidth=1,linestyle='dotted')
        pylab.imshow(M.transpose(),cmap=cmap,origin="lower",aspect="auto")
        pylab.title(title)
        #pylab.draw()

    def plot_autocorrelation(self,coordinates, max_lag=None, figure=None, alpha_line=0.8, color_line=None, title="Autocorrelation"): 
        autocorrelation = self._tracer.autocorrelation(coordinates) 
        if autocorrelation  is None: 
            return False
        else: 
            if color_line  is None: 
                color_line = hsv_to_rgb(0.0,0.85,1)
            if not figure:
                pass #figure = pylab.figure()
            else: 
                pylab.figure(figure.number)
            if max_lag is None: 
                max_lag = self._tracer.size()
            elif max_lag>=self._tracer.size() or max_lag<=0:
                max_lag = self._tracer.size()
            pylab.plot(autocorrelation[0:max_lag],alpha=alpha_line,color=color_line) 

            pylab.title(title)
            #pylab.draw()
        return figure

    def plot_acceptance_ratio(self,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title="Acceptance Ratio"): 
        return self._plot_simple_curves(self._tracer.get_acceptance_ratio(),figure, alpha_line, alpha_fill, color_line, color_fill,title) 

    def plot_accepted(self,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title="Accepted"): 
        return self._plot_simple_curves(self._tracer.get_accepted(),figure, alpha_line, alpha_fill, color_line, color_fill,title) 

    def plot_threshold(self,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title="Threshold"): 
        return self._plot_simple_curves(self._tracer.get_threshold(),figure, alpha_line, alpha_fill, color_line, color_fill,title) 

    def _plot_simple_curves(self,curve,figure=None, alpha_line=0.8, alpha_fill=0.2, color_line=None, color_fill=None,title=" "): 
        if color_line  is None: 
            color_line = hsv_to_rgb(0.0,0.85,1)
        if color_fill  is None: 
            color_fill = hsv_to_rgb(0.0,0.85,1)
        if not figure:
            pass #figure = pylab.figure()
        else: 
            pylab.figure(figure.number)
        pylab.plot(curve,alpha=alpha_line,color=color_line)
        pylab.title(title)
        #pylab.draw()
        return figure



# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


from ilang.Graphs import Dependence, ProbabilisticGraphicalModel
from ilang.Models import MultivariateGaussian 
from ilang.Samplers import Sampler
from ilang.Tracers import RamTracer
from ilang.Display import MatplotlibDisplay
import numpy




# Define the model 
ndim = 10
model = MultivariateGaussian('gaussian') 

# Build the graph 
dag = ProbabilisticGraphicalModel(['x','mu','cov']) 
dag.set_nodes_given(['mu','cov'], True) 
dag.add_dependence(model,{'x':'x','mu':'mu','cov':'cov'}) 

# Initialize the nodes of the graph
dag.set_node_value('x',numpy.ones((1,ndim))) 
dag.set_node_value('mu',numpy.zeros((1,ndim)))
dag.set_node_value('cov',numpy.eye(ndim)) 

# Initialise sampler, tracer, display 
sampler = Sampler(dag) 
tracer = RamTracer(sampler)   
display = MatplotlibDisplay(tracer) 

# Sample 
sampler.sample(1000,trace=False) 
#display.plot('mu') 
#display.plot('sigma') 


if __name__=="__main__": 
    dag.webdisplay(background=False) 


